# pylint: disable=django-not-available,invalid-name,import-outside-toplevel,too-many-lines
import base64
import codecs
import io
import logging
import os
import pickle
import random
import subprocess
import sys
import tempfile
import time
import json
import traceback
import types
import typing as tp
import uuid
import warnings
from functools import lru_cache, wraps
from pathlib import Path
from threading import Thread
from enum import Enum

from django.conf import settings
from django.middleware.gzip import GZipMiddleware
from django.urls import re_path, include
from django.core.mail import get_connection, EmailMultiAlternatives
from django.core.cache import caches, InvalidCacheBackendError
from django.core.paginator import Paginator as BasePaginator
from django.template import loader
from django.utils import translation, functional
from django.utils.translation import get_language
from django.utils.module_loading import import_string as import_class
from django.views.decorators.csrf import csrf_exempt
from django.views.generic.base import View

from . import exceptions as ex
from .tools import multikeysort


logger: logging.Logger = logging.getLogger('vstutils')
ON_POSIX = 'posix' in sys.builtin_module_names


def deprecated(func: tp.Callable):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""

    def new_func(*args, **kwargs):
        warnings.warn(f'Call to deprecated function {func.__name__}.',
                      category=DeprecationWarning,
                      stacklevel=2)
        return func(*args, **kwargs)

    return new_func


def is_member_descriptor(obj):
    try:
        return type(obj).__name__ == 'member_descriptor'
    except:  # nocv
        return False


@lru_cache()
def current_lang(lang=None):
    from vstutils.api.models import Language
    try:
        return Language.objects.get(code=lang or get_language())
    except Exception:  # nocv
        return Language.objects.all().first()


def get_render(name: tp.Text, data: tp.Dict, trans: tp.Text = 'en') -> tp.Text:
    """
    Render string from template.

    :param name: -- full template name
    :type name: str
    :param data: -- dict of rendered vars
    :type data: dict
    :param trans: -- translation for render. Default 'en'.
    :type trans: str
    :return: -- rendered string
    :rtype: str
    """
    cur_language = translation.get_language()
    try:
        if trans != cur_language:
            translation.activate(trans)
        config = loader.get_template(name)
        result = config.render(data, data.pop('request', None)).replace('\r', '')
    finally:
        if trans != cur_language:
            translation.activate(cur_language)
    return result


def encode(key, clear):
    """
    Encode string by Vigenère cipher.

    :param key: -- secret key for encoding
    :type key: str
    :param clear: -- clear value  for encoding
    :type clear: str
    :return: -- encoded string
    :rtype: str
    """
    # pylint: disable=consider-using-enumerate

    enc = []
    for i in range(len(clear)):
        key_c = key[i % len(key)]
        enc.append(chr((ord(clear[i]) + ord(key_c)) % 256))
    return base64.urlsafe_b64encode("".join(enc).encode()).decode()


def decode(key, enc):
    """
    Decode string from encoded by Vigenère cipher.

    :param key: -- secret key for encoding
    :type key: str
    :param enc: -- encoded string for decoding
    :type enc: str
    :return: -- decoded string
    :rtype: str
    """
    # pylint: disable=consider-using-enumerate

    dec = []
    enc = base64.urlsafe_b64decode(enc).decode()
    for i in range(len(enc)):
        key_c = key[i % len(key)]
        dec.append(chr((256 + ord(enc[i]) - ord(key_c)) % 256))
    return "".join(dec)


def get_if_lazy(obj):
    with raise_context():
        if isinstance(obj, functional.LazyObject):
            # pylint: disable=protected-access
            obj._setup() if obj._wrapped == functional.empty else None
            return obj._wrapped
    return obj


def send_mail(subject, message, from_email, recipient_list,  # noqa: CFQ002
              fail_silently=False, auth_user=None, auth_password=None,
              connection=None, html_message=None, **kwargs):
    """
    Wrapper over :func:`django.core.mail.send_mail` which provide additional named arguments.
    """
    # pylint: disable=too-many-arguments
    connection = connection or get_connection(
        username=auth_user,
        password=auth_password,
        fail_silently=fail_silently,
    )
    mail = EmailMultiAlternatives(
        subject,
        message,
        from_email,
        recipient_list,
        connection=connection,
        **kwargs,
    )
    if html_message:
        mail.attach_alternative(html_message, 'text/html')

    return mail.send()


def send_template_email_handler(
        subject: tp.Text,
        email_from: tp.Text,
        email: tp.Union[tp.List, tp.Text],
        template_name: tp.Text,
        context_data: tp.Optional[tp.Dict] = None,
        **kwargs,
) -> tp.SupportsInt:
    """
    Function for email sending.
    The function convert recipient to list and set context before sending if it possible.

    :param subject: mail subject.
    :param email_from: sender that be setup in email.
    :param email: list of strings or single string, with email addresses of recipients
    :param template_name: relative path to template in `templates` directory, must include extension in file name.
    :param context_data: dictionary with context for rendering message template.
    :param kwargs: additional named arguments for `send_mail`
    :return: Number of emails sent.
    """
    recipient_list = email if isinstance(email, (list, tuple)) else [email]

    if context_data is None:
        context = {}
    elif not isinstance(context_data, dict):
        context = dict(context_data)
    else:
        context = context_data.copy()

    return send_mail(
        subject=subject,
        message="",
        from_email=email_from,
        recipient_list=recipient_list,
        html_message=loader.render_to_string(
            template_name,
            context=context,
            request=context.pop('request', None)
        ),
        **kwargs
    )


def send_template_email(sync: bool = False, **kwargs):
    """
    Function executing sync or async email sending; according `sync` argument and settings variable "RPC_ENABLED".
    You can use this function to send message, it sends message asynchronously or synchronously.
    If you don't set settings for celery or don't have celery it sends synchronously mail.
    If celery is installed and configured and `sync` argument of the function is set to `False`,
    it sends asynchronously email.

    :param sync: argument for determining how send email, asynchronously or synchronously
    :param subject: mail subject.
    :param email: list of strings or single string, with email addresses of recipients
    :param template_name: relative path to template in `templates` directory, must include extension in file name.
    :param context_data: dictionary with context for rendering message template.
    """
    if sync or not settings.RPC_ENABLED:
        send_template_email_handler(email_from=settings.EMAIL_FROM_ADDRESS, **kwargs)
    else:
        from .tasks import SendEmailMessage
        SendEmailMessage.do(email_from=settings.EMAIL_FROM_ADDRESS, **kwargs)


def patch_gzip_response(response, request):
    if not response.status_code == 200:
        return  # nocv
    GZipMiddleware.process_response(None, request, response)


def patch_gzip_response_decorator(func):
    def gzip_response_wrapper(view, request, *args, **kwargs):
        response = func(view, request, *args, **kwargs)
        with raise_context():
            patch_gzip_response(response, request)
        return response

    return gzip_response_wrapper


def translate(text: tp.Text) -> tp.Text:
    """
    The ``translate`` function supports translation message dynamically
    with standard i18n vstutils'es mechanisms usage.

    Uses :func:`django.utils.translation.get_language` to get the language code and
    tries to get the translation from the list of available ones.

    :param text: Text message which should be translated.
    """
    try:
        return current_lang(lang=get_language()).translate(text)
    except Exception:  # nocv
        return text


def lazy_translate(text: tp.Text) -> functional.Promise:
    """
    The ``lazy_translate`` function has the same behavior as :func:`.translate`, but wraps it in a lazy promise.

    This is very useful, for example, for translating error messages in
    class attributes before the language code is known.

    :param text: Text message which should be translated.
    """
    return functional.lazy(translate, str)(text)


def create_view(model, **meta_options):
    """
    A simple function for getting the generated view by standard means, but with overloaded meta-parameters.
    This method can completely get rid of the creation of proxy models.

    Example:
        .. sourcecode:: python

            from vstutils.utils import create_view

            from .models import Host

            # Host model has full :class:`vstutils.api.base.ModelViewSet` view.
            # For overriding and create simple list view just setup this:
            HostListViewSet = create_view(
                HostList,
                view_class='list_only'
            )


    .. note::
        This method is also recommended in cases where there is a problem of recursive imports.


    :type model: Type[vstutils.models.BaseModel]
    :param model: Model class with `.get_view_class` method. This method also has :class:`vstutils.models.BModel`.
    :rtype: vstutils.api.base.GenericViewSet
    """
    return model.get_view_class(**meta_options)


class apply_decorators:
    """
    Decorator which apply list of decorators on method or class.

    Example:
        .. sourcecode:: python

            from vstutils.utils import apply_decorators

            def decorator_one(func):
                print(f"Decorated {func.__name__} by first decorator.")
                return func

            def decorator_two(func):
                print(f"Decorated {func.__name__} by second decorator.")
                return func

            @apply_decorators(decorator_one, decorator_two)
            def decorated_function():
                # Function decorated by both decorators.
                print("Function call.")
    """

    __slots__ = ('decorators',)

    def __init__(self, *decorators):
        self.decorators = decorators

    def __call__(self, func):
        for decorator in self.decorators:
            func = decorator(func)
        return func


class ClassPropertyMeta(type):
    def __setattr__(cls, key, value):
        obj = cls.__dict__.get(key, None)
        if isinstance(obj, ClassPropertyDescriptor):
            return obj.__set__(cls, value)
        return super().__setattr__(key, value)


class ClassPropertyDescriptor:
    __slots__ = ('fget', 'fset')

    meta = ClassPropertyMeta

    def __init__(self, fget: tp.Callable, fset: tp.Callable = None):
        self.fget, self.fset = self._fix_function(fget), self._fix_function(fset)

    def __get__(self, obj, klass=None):
        if obj is not None:
            return self.fget.__get__(obj, obj)()
        return self.fget.__get__(obj, type(obj) if klass is None else klass)()

    def __set__(self, obj: tp.Any, value: tp.Any):
        if not self.fset:
            raise AttributeError("can't set attribute")
        if obj is not None:
            return self.fset.__get__(obj, obj)(value)
        return self.fset.__get__(obj, type(obj))(value)  # nocv

    def setter(self, func: tp.Union[tp.Callable, classmethod]):
        self.fset = self._fix_function(func)
        return self

    @classmethod
    def _fix_function(cls, func):
        if func is None:
            return func
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        return func


class classproperty(ClassPropertyDescriptor):
    """
    Decorator which makes class method as class property.

    Example:
        .. sourcecode:: python

            from vstutils.utils import classproperty

            class SomeClass(metaclass=classproperty.meta):
                # Metaclass is needed for set attrs in class
                # instead of and not only object.

                some_value = None

                @classproperty
                def value(cls):
                    return cls.some_value

                @value.setter
                def value(cls, new_value):
                    cls.some_value = new_value
    """


class redirect_stdany:
    """
    Context for redirect any output to own stream.

    .. note::
        - On context return stream object.
        - On exit return old streams
    """
    __slots__ = ('stream', 'streams', '_old_streams')

    _streams: tp.ClassVar[tp.List[tp.Text]] = ["stdout", "stderr"]

    def __init__(self, new_stream=io.StringIO(), streams=None):
        """
        :param new_stream: -- stream where redirects all
        :type new_stream: object
        :param streams: -- names of streams like ``['stdout', 'stderr']``
        :type streams: list
        """
        self.streams = streams or self._streams
        self.stream = new_stream
        self._old_streams = {}

    def __enter__(self):
        for stream in self.streams:
            self._old_streams[stream] = getattr(sys, stream)
            setattr(sys, stream, self.stream)
        return self.stream

    def __exit__(self, exctype, excinst, exctb):
        for stream in self.streams:
            setattr(sys, stream, self._old_streams.pop(stream))


class Dict(dict):
    """
    Wrapper over `dict` which
    return JSON on conversion to string.
    """

    def __repr__(self):  # nocv
        return self.__str__()

    def __str__(self):
        return json.dumps(self.copy())


class tmp_file:
    """
    Temporary file with name
    generated and auto removed on close.

    **Attributes**:

    :param data: -- string to write in tmp file.
    :type data: str
    :param mode: -- file open mode. Default 'w'.
    :type mode: str
    :param bufsize: -- bufer size for tempfile.NamedTemporaryFile
    :type bufsize: int
    :param kwargs:  -- other kwargs for tempfile.NamedTemporaryFile

    """
    __slots__ = ('fd', 'path')

    def __init__(self, data: tp.Text = "", mode: tp.Text = "w", bufsize: int = -1, **kwargs):
        # pylint: disable=consider-using-with
        self.fd = tempfile.NamedTemporaryFile(mode, buffering=bufsize, **kwargs)
        self.path = Path(self.fd.name)
        if data:
            self.write(data)

    def write(self, wr_string: tp.Text):
        """
        Write to file and flush

        :param wr_string: -- writable string
        :type wr_string: str
        :return: None
        :rtype: None
        """
        result = self.fd.write(wr_string)
        self.fd.flush()
        return result

    def __getattr__(self, name: tp.Text):
        return getattr(self.fd, name)

    def __del__(self):
        with raise_context():
            self.fd.close()

    def __enter__(self):
        """
        :return: -- file object
        :rtype: tempfile.NamedTemporaryFile
        """
        return self

    def __exit__(self, type_e, value, tb):
        self.fd.close()


class tmp_file_context:
    """
    Context object for work with tmp_file.
    Auto close on exit from context and
    remove if file still exist.

    This context manager over :class:`.tmp_file`
    """
    __slots__ = ('tmp',)

    def __init__(self, *args, **kwargs):
        self.tmp: tempfile.NamedTemporaryFile = tmp_file(*args, **kwargs)

    def __enter__(self):
        return self.tmp

    def __exit__(self, type_e, value, tb):
        self.tmp.close()
        if self.tmp.path.exists():
            self.tmp.path.unlink()


class assertRaises:
    __slots__ = ('_kwargs', '_verbose', '_exclude', '_excepts', '_log', '__failed')

    def __init__(self, *args, **kwargs):
        """
        :param args: -- list of exception classes should be passed
        :type args: list,Exception
        :param exclude: -- list of exception classes should be raised
        :type exclude: list,Exception
        :param verbose: -- logging
        :type verbose: bool
        """
        self._kwargs = dict(**kwargs)
        self._verbose = kwargs.pop("verbose", settings.DEBUG)
        self._exclude = kwargs.pop("exclude", False)
        self._excepts = tuple(args)
        self.__failed = False

    def mark_as_failed(self):
        self.__failed = True

    def cleanup_fails(self):
        self.__failed = False

    @property
    def is_failed(self):
        return self.__failed

    def __enter__(self):
        return self  # pragma: no cover

    def __exit__(self, exc_type, exc_val, exc_tb):
        return exc_type is not None and (
                (not self._exclude and not issubclass(exc_type, self._excepts)) or
                (self._exclude and issubclass(exc_type, self._excepts))
        )


# noinspection PyUnreachableCode
class raise_context(assertRaises):
    """
    Context for exclude exceptions.
    """

    __slots__ = ()

    def execute(self, func: tp.Callable, *args, **kwargs):
        self.cleanup_fails()
        with self.__class__(self._excepts, **self._kwargs):
            try:
                return func(*args, **kwargs)
            except:
                self.mark_as_failed()
                type, value, traceback_obj = sys.exc_info()
                if self._verbose:
                    logger.debug(traceback.format_exc())
                raise
        return type, value, traceback_obj

    def __enter__(self):
        return self.execute

    def __call__(self, original_function: tp.Callable):
        def wrapper(*args, **kwargs):
            return self.execute(original_function, *args, **kwargs)
        return wrapper


class raise_context_decorator_with_default(raise_context):
    """
    Context for exclude errors and return default value.

    Example:
        .. sourcecode:: python

            from yaml import load
            from vstutils.utils import raise_context_decorator_with_default


            @raise_context_decorator_with_default(default={})
            def get_host_data(yaml_path, host):
                with open(yaml_path, 'r') as fd:
                    data = load(fd.read(), Loader=Loader)
                return data[host]
                # This decorator used when you must return some value even on error
                # In log you also can see traceback for error if it occur

            def clone_host_data(host):
                bs_data = get_host_data('inventories/aws/hosts.yml', 'build_server')
                ...

    """
    __slots__ = ('default_value',)

    def __init__(self, *args, **kwargs):
        self.default_value = kwargs.pop('default', None)
        super().__init__(*args, **kwargs)

    def execute(self, func: tp.Callable, *args, **kwargs):
        result = super().execute(func, *args, **kwargs)
        if self.is_failed:
            return self.default_value
        return result


class exception_with_traceback(raise_context):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val is not None:
            exc_val.traceback = traceback.format_exc()
            raise exc_val.with_traceback(exc_tb)


class BaseVstObject:
    """
    Default mixin-class for custom objects which needed to get settings and cache.
    """
    __slots__ = ()

    @classmethod
    def get_django_settings(cls, name: tp.Text, default: tp.Any = None):
        # pylint: disable=access-member-before-definition
        """
        Get params from Django settings.

        :param name: name of param
        :type name: str
        :param default: default value of param
        :type default: object
        :return: Param from Django settings or default.
        """
        return getattr(settings, name, default)

    @classmethod
    def get_django_cache(cls, cache_name: tp.Text = 'default'):
        try:
            return caches[cache_name]
        except InvalidCacheBackendError:  # nocv
            return caches['default']


class SecurePickling(BaseVstObject):
    """
    Secured pickle wrapper by Vigenère cipher.


    .. warning::
        Do not use it with untrusted transport anyway.

    Example:
        .. sourcecode:: python

            from vstutils.utils import SecurePickling


            serializer = SecurePickling('password')

            # Init secret object
            a = {"key": "value"}
            # Serialize object with secret key
            pickled = serializer.dumps(a)
            # Deserialize object
            unpickled = serializer.loads(pickled)

            # Check, that object is correct
            assert a == unpickled

    """
    __slots__ = ('secure_key',)

    def __init__(self, secure_key: tp.Optional[tp.Text] = None):
        """
        :param secure_key: Secret key for encoding.
        """
        if secure_key is None:
            secure_key = self.get_django_settings('SECRET_KEY')
        self.secure_key = str(secure_key)

    def _encode(self, value: tp.Text):
        return encode(self.secure_key, value)

    def _decode(self, value: tp.Text):
        return decode(self.secure_key, value)

    def loads(self, value: tp.Any):
        return pickle.loads(codecs.decode(self._decode(value).encode(), "base64"))  # nosec

    def dumps(self, value: tp.Any):
        return self._encode(codecs.encode(pickle.dumps(value), "base64").decode())


class Executor(BaseVstObject):
    """
    Command executor with realtime output write and line handling.
    By default and by design executor initialize string attribute ``output``
    which will be modifyed by ``+=`` operator with new lines by :meth:`.Executor.write_output`
    procedure. Override the method if you want change behavior.

    Executor class supports periodically (0.01 sec) handling process and execute some checks
    by overriding :meth:`.Executor.working_handler` procedure method. If you want disable this behavior
    override the method by None value or use :class:`.UnhandledExecutor`.
    """
    __slots__ = ('output', '__stdout__', '__stderr__', 'env')

    CANCEL_PREFIX: tp.ClassVar[tp.Text] = "CANCEL_EXECUTE_"
    newlines: tp.ClassVar[tp.List[tp.Text]] = ['\n', '\r\n', '\r']
    STDOUT: tp.ClassVar[int] = subprocess.PIPE
    STDERR: tp.ClassVar[int] = subprocess.STDOUT
    DEVNULL: tp.ClassVar[int] = subprocess.DEVNULL
    CalledProcessError: tp.ClassVar[tp.Type[subprocess.CalledProcessError]] = subprocess.CalledProcessError

    env: tp.Dict[tp.Text, tp.Text]

    def __init__(self, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, **environ_variables):
        """
        :type stdout: BinaryIO,int
        :type stderr: BinaryIO,int
        """
        self.output = ''
        self.__stdout__ = stdout
        self.__stderr__ = stderr
        self.env = environ_variables

    def write_output(self, line: tp.Text) -> None:
        """
        :param line: -- line from command output
        :type line: str
        :return: None
        :rtype: None
        """
        self.output += str(line)

    def _handle_process(self, proc: subprocess.Popen, stream: tp.Text):
        stream_object = getattr(proc, stream)
        while not stream_object.closed:
            self.working_handler(proc)
            time.sleep(0.01)

    def working_handler(self, proc: subprocess.Popen) -> None:
        # pylint: disable=unused-argument
        """
        Additional handler for executions.

        :type proc: subprocess.Popen
        """

    def _unbuffered(self, proc: subprocess.Popen, stream: tp.Text = 'stdout'):
        """
        Unbuffered output handler.

        :type proc: subprocess.Popen
        :type stream: str
        :return:
        """
        if self.working_handler is not None:
            t = Thread(target=self._handle_process, args=(proc, stream))
            t.start()
        out = getattr(proc, stream)
        try:
            retcode = None
            while retcode is None:
                for line in iter(out.readline, ""):
                    yield line.rstrip()
                retcode = proc.poll()
        finally:
            out.close()

    def line_handler(self, line: tp.Text) -> None:
        if line is not None:  # pragma: no branch
            with raise_context():
                self.write_output(line)

    def execute(self, cmd: tp.List[tp.Text], cwd: tp.Union[tp.Text, Path]) -> tp.Text:
        """
        Execute commands and output this.

        :param cmd: -- list of cmd command and arguments
        :type cmd: list
        :param cwd: -- workdir for executions
        :type cwd: str
        :return: -- string with full output
        :rtype: str
        """
        self.output = ""
        env = os.environ.copy()
        env.update(self.env)
        with subprocess.Popen(
            cmd, stdout=self.__stdout__, stderr=self.__stderr__,
            bufsize=0, universal_newlines=True,
            cwd=str(cwd), env=env,
            close_fds=ON_POSIX
        ) as proc:
            for line in self._unbuffered(proc):
                self.line_handler(line)
            return_code = proc.poll()
            if return_code:
                logger.error(self.output)
                raise subprocess.CalledProcessError(
                    return_code, cmd, output=str(self.output)
                )
        return self.output


class UnhandledExecutor(Executor):
    """
    Class based on :class:`.Executor` but disables `working_handler`.
    """
    __slots__ = ()
    working_handler = None  # type: ignore


class KVExchanger(BaseVstObject):
    """
    Class for transmit data using key-value fast (cache-like) storage between
    services. Uses same cache-backend as Lock.
    """
    __slots__ = ('key', 'timeout', '__djangocache__')
    TIMEOUT: tp.ClassVar[int] = 60

    @classproperty
    def PREFIX(cls):
        # pylint: disable=no-self-argument
        return f"{cls.get_django_settings('VST_PROJECT_LIB')}_exchange_"

    @classproperty
    def cache(cls):
        # pylint: disable=no-self-argument,no-member,access-member-before-definition
        if hasattr(cls, '__djangocache__') and not is_member_descriptor(cls.__djangocache__):
            return cls.__djangocache__
        cls.__djangocache__ = cls.get_django_cache('locks')
        return cls.cache

    def __init__(self, key, timeout=None):
        self.key = self.PREFIX + str(key)
        self.timeout = timeout or self.TIMEOUT

    def send(self, value, ttl=None):
        # pylint: disable=no-member
        return self.cache.add(self.key, value, ttl or self.timeout)

    def prolong(self, ttl=None):
        # pylint: disable=no-member
        self.cache.touch(self.key, ttl or self.timeout)

    def get(self):
        # pylint: disable=no-member
        value = self.cache.get(self.key)
        self.cache.delete(self.key)
        return value

    def delete(self):
        # pylint: disable=no-member
        self.cache.delete(self.key)


class Lock(KVExchanger):
    """
    Lock class for multi-jobs workflow. Based on :class:`.KVExchanger`.
    The Lock allows only one thread to enter the part that's locked and shared
    between apps using one locks cache (see also `[locks] <config.html#locks-settings>`_).


    :param id: -- unique id for lock.
    :type id: int,str
    :param payload: -- lock additional info. Should be any boolean True value.
    :param repeat: -- time to wait lock.release. Default 1 sec.
    :type repeat: int
    :param err_msg: -- message for AcquireLockException error.
    :type err_msg: str

    .. note::
        - Used django.core.cache lib and settings in `settings.py`
        - Have Lock.SCHEDULER and Lock.GLOBAL id

    Example:
        .. sourcecode:: python

            from vstutils.utils import Lock

            with Lock("some_lock_identifier", repeat=30, err_msg="Locked by another proccess") as lock:
                # where
                # ``"some_lock_identifier"`` is unique id for lock and
                # ``30`` seconds lock is going wait until another process will release lock id.
                # After 30 sec waiting lock will raised with :class:`.Lock.AcquireLockException`
                # and ``err_msg`` value as text.
                some_code_execution()
                # ``lock`` object will has been automatically released after
                # exiting from context.

    Another example without context manager:
        .. sourcecode:: python

            from vstutils.utils import Lock

            # locked block after locked object created
            lock = Lock("some_lock_identifier", repeat=30, err_msg="Locked by another proccess")
            # deleting of object calls ``lock.release()`` which release and remove lock from id.
            del lock

    """
    __slots__ = ('id', 'payload_data')
    TIMEOUT: tp.ClassVar[int] = 60 * 60 * 24
    GLOBAL: tp.ClassVar[tp.Text] = "global-deploy"
    SCHEDULER: tp.ClassVar[tp.Text] = "celery-beat"

    class AcquireLockException(Exception):
        """ Exception which will raised on unreleased lock. """

    @classproperty
    def PREFIX(cls):
        # pylint: disable=no-self-argument
        return f"{cls.get_django_settings('VST_PROJECT_LIB')}_lock_"

    def __init__(self, id, payload=1, repeat=1, err_msg="", timeout=None):
        # pylint: disable=too-many-arguments
        super().__init__(id, timeout)
        self.payload_data = f'{uuid.uuid4()}_{payload}'
        self.id, start = None, time.time()
        while time.time() - start <= repeat:
            if self.send(self.payload_data) and not time.sleep(random.random()) and self.get() == self.payload_data:
                logger.debug(f'Acquire lock with id `{id}` and payload `{self.payload_data}`')
                self.id = id
                return
            time.sleep(random.random() / 10)
        raise self.AcquireLockException(err_msg)

    def get(self):  # nocv
        # pylint: disable=no-member
        return self.cache.get(self.key)

    def __enter__(self):
        return self

    def __exit__(self, type_e, value, tb):
        self.release()

    def release(self, force_release=False):
        # pylint: disable=no-member
        if force_release or self.id is not None:
            self.cache.delete(self.key)

    def __del__(self):
        self.release()


class __LockAbstractDecorator:
    __slots__ = ('kwargs',)
    _err = "Wait until the end."
    _lock_key = None

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.kwargs["err_msg"] = self.kwargs.get("err_msg", self._err)

    def execute(self, func, *args, **kwargs):
        if self._lock_key is not None:
            with Lock(self._lock_key, **self.kwargs):
                return func(*args, **kwargs)
        return func(*args, **kwargs)

    def __call__(self, original_function):
        @wraps(original_function)
        def wrapper(*args, **kwargs):
            return self.execute(original_function, *args, **kwargs)

        return wrapper


class model_lock_decorator(__LockAbstractDecorator):
    """
    Decorator for functions where 'pk' kwarg exist
    for lock by id.

    .. warning::
        - On locked error raised ``Lock.AcquireLockException``
        - Method must have and called with ``pk`` named arg.

    """
    __slots__ = ('_lock_key',)
    _err = "Object locked. Wait until unlock."

    def execute(self, func, *args, **kwargs):
        self._lock_key = kwargs.get('pk', None)
        return super().execute(func, *args, **kwargs)


class Paginator(BasePaginator):
    """
    Class for fragmenting the query for small queries.
    """
    __slots__ = ()

    def __init__(self, qs, chunk_size=None):
        """
        :param qs: -- queryset for fragmenting
        :type qs: django.db.models.QuerySet
        :param chunk_size: -- size of the fragments.
        :type chunk_size: int
        """
        chunk_size = chunk_size or BaseVstObject().get_django_settings("PAGE_LIMIT", None)
        super().__init__(qs, chunk_size)

    def __iter__(self):
        for page in range(1, self.num_pages + 1):
            yield self.page(page)

    def items(self):
        for page in self:
            for obj in page.object_list:
                obj.paginator = self
                obj.page = page
                yield obj


class ObjectHandlers(BaseVstObject):
    """
    Handlers wrapper for get objects from some settings structure.

    Example:
        .. sourcecode:: python

            from vstutils.utils import ObjectHandlers

            '''
            In `settings.py` you should write some structure:

            SOME_HANDLERS = {
                "one": {
                    "BACKEND": "full.python.path.to.module.SomeClass"
                },
                "two": {
                    "BACKEND": "full.python.path.to.module.SomeAnotherClass",
                    "OPTIONS": {
                        "some_named_arg": "value"
                    }
                }
            }
            '''

            handlers = ObjectHandlers('SOME_HANDLERS')

            # Get class handler for 'one'
            one_backend_class = handlers['one']
            # Get object of backend 'two'
            two_obj = handlers.get_object()
            # Get object of backend 'two' with overriding constructor named arg
            two_obj_overrided = handlers.get_object(some_named_arg='another_value')

    """

    __slots__ = ('type', 'err_message', '__list__', '__loaded_backends__')

    type: tp.Text
    err_message: tp.Optional[tp.Text]

    def __init__(self, type_name: tp.Text, err_message: tp.Text = None):
        """
        :param type_name: -- type name for backends.Like name in dict.
        :type type_name: str
        """
        self.type = type_name
        self.err_message = err_message
        self.__list__: tp.Optional[tp.Dict[tp.Text, tp.Any]] = None
        self.__loaded_backends__: tp.Dict[tp.Text, tp.Any] = {}

    @property
    def objects(self):
        return {name: self[name] for name in self.list()}

    def __len__(self):  # pragma: no cover
        return len(self.objects)

    def __iter__(self):
        return iter(self.items())

    def __getitem__(self, name: tp.Text) -> tp.Any:
        return self.backend(name)

    def __call__(self, name, obj):
        return self.get_object(name, obj)

    def __dict__(self):  # pragma: no cover
        return self.items()

    def keys(self) -> tp.Iterable[tp.Text]:
        return self.objects.keys()

    def values(self) -> tp.Iterable:  # pragma: no cover
        return dict(self).values()

    def items(self):
        return self.objects.items()

    def list(self) -> tp.Dict[tp.Text, tp.Dict[tp.Text, tp.Any]]:
        if self.__list__ is None:
            self.__list__ = self.get_django_settings(self.type, {})
        return self.__list__

    def _get_backend(self, backend):
        if backend in self.__loaded_backends__:
            return self.__loaded_backends__[backend]
        self.__loaded_backends__[backend] = import_class(backend)
        return self.__loaded_backends__[backend]

    def get_backend_data(self, name: tp.Text):
        return self.list()[name]

    def get_backend_handler_path(self, name: tp.Text):
        return self.get_backend_data(name).get('BACKEND', None)

    def backend(self, name: tp.Text):
        """
        Get backend class

        :param name: -- name of backend type
        :type name: str
        :return: class of backend
        :rtype: type,types.ModuleType,object
        """
        try:
            backend = self.get_backend_handler_path(name)
            if backend is None:
                raise ex.VSTUtilsException("Backend is 'None'.")  # pragma: no cover
            return self._get_backend(backend)
        except KeyError or ImportError as err:
            msg = f"{name} ({self.err_message})" if self.err_message else name
            raise ex.UnknownTypeException(msg) from err

    def opts(self, name: tp.Text):
        return self.get_backend_data(name).get('OPTIONS', {})

    def get_object(self, name: tp.Text, *args, **kwargs):
        opts = self.opts(name)
        opts.update(kwargs)
        return self[name](*args, **opts)


class StaticFilesHandlers(ObjectHandlers):
    def opts(self, name):
        opts = super().opts(name)
        opts['name'] = name
        return opts

    def get_static_objects(self):
        for name in self.keys():
            for result in self.get_object(name).spa_static_list:
                yield result

    def get_sorted_list(self):
        return tuple(multikeysort(self.get_static_objects(), ['priority']))


class ModelHandlers(ObjectHandlers):
    """
    Handlers for some models like 'INTEGRATIONS' or 'REPO_BACKENDS'.
    Based on :class:`.ObjectHandlers` but more specific for working with models.
    All handlers backends get by first argument model object.

    **Attributes**:

    :param objects: -- dict of objects like: ``{<name>: <backend_class>}``
    :type objects: dict
    :param keys: -- names of supported backends
    :type keys: list
    :param values: -- supported backends classes
    :type values: list
    """

    __slots__ = ()

    def get_object(self, name: tp.Text, obj) -> tp.Any:  # type: ignore
        """
        :param name: -- string name of backend
        :param name: str
        :param obj: -- model object
        :type obj: django.db.models.Model
        :return: backend object
        :rtype: object
        """
        return self[name](obj, **self.opts(name))


class URLHandlers(ObjectHandlers):
    """
    Object handler for GUI views. Uses `GUI_VIEWS` from settings.py.
    Based on :class:`.ObjectHandlers` but more specific to urlpatterns.

    Example:
        .. sourcecode:: python

            from vstutils.utils import URLHandlers


            # By default gets from `GUI_VIEWS` in `settings.py`
            urlpatterns = list(URLHandlers())
    """
    __slots__ = ('additional_handlers', '__handlers__', 'default_namespace')

    settings_urls: tp.ClassVar[tp.List[tp.Text]] = [
        'LOGIN_URL',
        'LOGOUT_URL'
    ]
    additional_handlers: tp.List[tp.Text]

    def __init__(self, type_name: tp.Text = 'URLS', *args, **kwargs):
        self.additional_handlers = kwargs.pop('additional_handlers', ['VIEWS']) + [type_name]
        self.default_namespace = kwargs.pop('namespace', None)
        self.__handlers__ = None
        super().__init__(type_name, *args, **kwargs)

    @property
    def view_handlers(self):
        if not self.__handlers__:
            self.__handlers__ = tuple(map(self.__class__, self.additional_handlers))
            # self.__handlers__ = []
            # handler_class = self.__class__
            # for handler_settings_name in self.additional_handlers:
            #     self.__handlers__.append(handler_class(handler_settings_name))
        return self.__handlers__

    def get_backend_data(self, name: tp.Text):
        data = super().get_backend_data(name)
        if isinstance(data, str):
            for handler in self.view_handlers:
                try:
                    return handler.get_backend_data(data)
                except:  # nosec
                    continue
            raise ex.VSTUtilsException(f'Invalid handler name "{data}"')  # nocv
        return data

    def get_object(self, name: tp.Text, *argv, **kwargs):
        """
        Get url object tuple for urls.py

        :param name: url regexp from
        :type name: str
        :param argv: overrided args
        :param kwargs: overrided kwargs
        :return: url object
        :rtype: django.urls.re_path
        """
        regexp = name
        options = self.opts(regexp)
        options.update(kwargs)
        args = options.pop('view_args', argv)
        view_kwargs = options.pop('view_kwargs', {})
        csrf_enable = self.get_backend_data(regexp).get('CSRF_ENABLE', True)
        if regexp in self.settings_urls:
            regexp = rf'^{self.get_django_settings(regexp)[1:]}'
        view_class = self[name]
        namespace = view_kwargs.pop('namespace', self.default_namespace)
        result: tp.Union[tp.Tuple, types.ModuleType]
        if isinstance(view_class, View) or hasattr(view_class, 'as_view'):
            view = view_class.as_view(**view_kwargs)
            if not csrf_enable:
                view = csrf_exempt(view)
            return re_path(regexp, view, *args, **options)
        elif (isinstance(view_class, types.ModuleType) and
              hasattr(view_class, 'urlpatterns') and
              hasattr(view_class, 'app_name')):
            result = view_class
            namespace = None
        else:
            result = (view_class, 'gui')
        return re_path(regexp, include(result, namespace=namespace), *args, **view_kwargs)

    def urls(self) -> tp.Iterable:
        for regexp in self.list():
            yield self.get_object(regexp)

    def __iter__(self):
        return self.urls()


def list_to_choices(items_list: tp.Iterable, response_type: tp.Callable = list) -> tp.Iterable[tp.Tuple[str, str]]:
    """
    Method to create django model choices from flat list of values.

    :param items_list: list of flat values.
    :param response_type: casting type of returned mapping
    :return: list of tuples from `items_list` values
    """
    return response_type(map(lambda x: (x, x), items_list))


class BaseEnum(str, Enum):
    """
    BaseEnum extends `Enum` class and used to create enum-like objects that can be used in django serializers or
    django models.

    Example:

        .. sourcecode:: python

            from vstutils.models import BModel

            class ItemCLasses(BaseEnum):
                FIRST='FIRST'
                SECOND='SECOND'
                THIRD='THIRD'


            class MyDjangoModel(BModel):
                item_class = models.CharField(max_length=1024, choices=ItemCLasses.to_choices())

                @property
                def is_second(self):
                    # Function check is item has second class of instance
                    return ItemCLasses.SECOND.is_equal(self.item_class)

    """

    def __new__(cls, name):
        return str.__new__(cls, name)

    def __repr__(self):
        return repr(self.__str__())

    def __str__(self):
        return str(self.name)

    def __hash__(self):
        return hash(str(self))

    @classmethod
    def get_names(cls):
        return tuple(x.name for x in cls)

    @classmethod
    def to_choices(cls):
        return list_to_choices(cls.get_names())

    def is_equal(self, cmp_str):
        return str(cmp_str) == str(self)

    def not_equal(self, cmp_str):
        return not self.is_equal(cmp_str)



"""
Default Django model classes overrides in `vstutils.models` module.
"""
import logging

from django.apps import apps
from django.db import models
from django.conf import settings

from .base import ModelBaseClass, get_proxy_labels, LAZY_MODEL
from .queryset import BQuerySet
from .model import BaseModel
from .decorators import register_view_action, register_view_method
from .fields import (
    NamedBinaryFileInJSONField,
    NamedBinaryImageInJSONField,
    MultipleNamedBinaryFileInJSONField,
    MultipleNamedBinaryImageInJSONField,
    FkModelField
)
from ..utils import raise_context


logger = logging.getLogger('vstutils')


class Manager(models.Manager.from_queryset(BQuerySet)):
    """
    Default VSTUtils manager. Used by `BaseModel` and `BModel`.
    Uses `BQuerySet` as base.
    """


class BModel(BaseModel):
    """
    Default model class that generates model viewset, separate serializers for list() and retrieve(), filters,
    api endpoints and nested views.

    Examples:
        .. sourcecode:: python

            from django.db import models
            from rest_framework.fields import ChoiceField
            from vstutils.models import BModel

            class Stage(BModel):
                name = models.CharField(max_length=256)
                order = models.IntegerField(default=0)

                class Meta:
                    default_related_name = "stage"
                    ordering = ('order', 'id',)
                    # fields which would be showed on list.
                    _list_fields = [
                        'id',
                        'name',
                    ]
                    # fields which would be showed on detail view and creation.
                    _detail_fields = [
                        'id',
                        'name',
                        'order'
                    ]
                    # make order as choices from 0 to 9
                    _override_detail_fields = {
                        'order': ChoiceField((str(i) for i in range(10)))
                    }


            class Task(BModel):
                name = models.CharField(max_length=256)
                stages = models.ManyToManyField(Stage)
                _translate_model = 'Task'

                class Meta:
                    # fields which would be showed.
                    _list_fields = [
                        'id',
                        'name',
                    ]
                    # create nested views from models
                    _nested = {
                        'stage': {
                            'allow_append': False,
                            'model': Stage
                        }
                    }


        In this case, you create models which could converted to simple view, where:

        - ``POST``/``GET`` to ``/api/version/task/`` - creates new or get list of tasks
        - ``PUT``/``PATCH``/``GET``/``DELETE`` to ``/api/version/task/:id/`` - updates, retrieves or removes instance of task
        - ``POST``/``GET`` to ``/api/version/task/:id/stage/`` - creates new or get list of stages in task
        - ``PUT``/``PATCH``/``GET``/``DELETE`` to ``/api/version/task/:id/stage/:stage_id`` - updates, retrieves or
          removes instance of stage in task.

        To attach a view to an API insert the following code in `settings.py`:

        .. sourcecode:: python

            API[VST_API_VERSION][r'task'] = {
                'model': 'your_application.models.Task'
            }

        For primary access to generated view inherit from `Task.generated_view` property.

        To make translation on frontend easier use ``_translate_model`` attribute with model_name

        List of meta-attributes for generating a view:

        - ``_view_class`` - list of additional view classes to inherit from,
          class or string to import with base class ViewSet.
          Constants are also supported:

            - ``read_only`` - to create a view only for viewing;
            - ``list_only`` - to create a view with list only;
            - ``history`` - to create a view only for viewing and deleting records.

          CRUD-view is applied by default.

        - ``_serializer_class`` - class of API serializer; use this attribute to specify parent class for autogenerated serializers.
          Default is :class:`vstutils.api.serializers.VSTSerializer`.
          Can take a string to import, serializer class or :class:`django.utils.functional.SimpleLazyObject`.
        - ``_serializer_class_name`` - model name for OpenAPI definitions.This would be a model name in generated admin interface.
          Default is name of model class.
        - ``_list_fields`` or ``_detail_fields`` - list of fields which will be listed in entity list or detail view accordingly.
          Same as DRF serializers meta-attribute "fields".
        - ``_override_list_fields`` or ``_override_detail_fields`` - mapping with names and field types
          that will be redeclared in serializer attributes(think of it as declaring fields in DRF ModelSerializer).
        - ``_properties_groups`` - dict with key as group name and value as list of fields(str). Allows to
          group fields in sections on frontend.
        - ``_view_field_name`` - name of field frontend shows as main view name.
        - ``_non_bulk_methods`` - list of methods which must not used via bulk requests.
        - ``_extra_serializer_classes`` - mapping with additional serializers in viewset. For example, custom serializer,
          which will compute smth in action (mapping name). Value can be string for import.
          Important note: setting `model` attribute to None allows to use standard serializer generation mechanism
          and get fields from a list or detail serializer (set `__inject_from__` serializer's meta attribute to
          `list` or `detail` accordingly). In some cases, it is required to pass the model to the serializer.
          For these purposes, the constant `LAZY_MODEL` can be used as a meta attribute. Each time the serializer is used,
          the exact model where this serializer was declared will be set.
        - ``_filterset_fields`` - list/dict of filterset names for API-filtering. Default is list of fields in list view.
          During processing a list of fields checks for the presence of special field names and inherit additional
          parent classes. If the list contains ``id``, class will inherit from
          :class:`vstutils.api.filters.DefaultIDFilter`. If the list contains ``name``, class will inherit from
          :class:`vstutils.api.filters.DefaultNameFilter`. If both conditions are present, inheritance will be from all
          of the above classes.
          Possible values include `list` of fields to filter or `dict` where key is a field name and value is
          a Filter class. Dict extends attribute functionality and provides ability to override filter field class
          (None value disables overriding).
        - ``_search_fields`` - tuple or list of fields using for search requests.
          By default (or `None`) get all filterable fields in detail view.
        - ``_copy_attrs`` - list of model-instance attributes indicates that object is copiable with this attrs.
        - ``_nested`` - key-value mapping with nested views (key - nested name,
          kwargs for :class:`vstutils.api.decorators.nested_view` decorator but supports
          ``model`` attribute as nested). ``model`` can be string for import.
        - ``_extra_view_attributes`` - key-value mapping with additional view attributes,
          but has less priority over generated attributes.


        In common, you can also add custom attributes to override or extend the default list of processing classes.
        Supported view attributes are ``filter_backends``, ``permission_classes``, ``authentication_classes``,
        ``throttle_classes``, ``renderer_classes`` and ``parser_classes``.
        List of meta-attributes for settings of view is looks like:

        - ``_pre_{attribute}`` - List of classes included before defaults.
        - ``_{attribute}`` - List of classes included after defaults.
        - ``_override_{attribute}`` - boolean flag indicates that attribute override default
          viewset (otherwise appends). Default is ``False``.

        .. note::
            You may need to create an
            `action <https://www.django-rest-framework.org/api-guide/viewsets/#marking-extra-actions-for-routing>`_
            on generated view. Use :class:`vstutils.models.decorators.register_view_action` decorator with the ``detail`` argument
            to determine applicability to a list or detail entry.
            In this case, the decorated method will take an instance of the view object as ``self`` attribute.

        .. note::
            In some cases, inheriting models may require to inherit Meta class from the base model.
            If the Meta is explicitly declared in the base class, then you can get it through
            the attribute `OriginalMeta` and use it for inheritance.

        .. note::
            Docstring of model will be reused for view descriptions. It is possible to write both a general description
            for all actions and description for each action using the following syntax:

            .. sourcecode::

                General description for all actions.

                action_name:
                    Description for this action.

                another_action:
                    Description for another action.
    """

    #: Primary field for select and search in API.
    id = models.AutoField(primary_key=True, max_length=20)
    #: If hidden is set to True, entry will be excluded from query in BQuerySet.
    hidden = models.BooleanField(default=False)
    hidden._hide = True  # pylint: disable=protected-access

    class Meta:
        abstract = True

    def __unicode__(self):
        return f"<{self.id}>"


@raise_context()
def bulk_notify_clients(channel=None, objects=()):
    if not settings.CENTRIFUGO_CLIENT_KWARGS:
        return
    notificator_class = apps.get_app_config('vstutils_api').module.notificator_class
    notificator = notificator_class([], channel=channel)
    for labels, pk in objects:
        notificator.create_notification(labels, pk)
    return notificator.send()


@raise_context()
def notify_clients(model, pk=None):
    logger.debug(f'Notify clients about model update: {model._meta.label}')
    bulk_notify_clients(objects=(
        ((model._meta.label, *get_proxy_labels(model)), pk),
    ))



#  pylint: disable=bad-super-call,unused-argument
from pathlib import Path

from markdown import markdown
from django.http.response import Http404
from django.contrib.auth.decorators import login_required
from django.views.generic.edit import FormView
from django.views.generic import TemplateView
from django.template.response import TemplateResponse
from django.contrib.auth import views as auth, login
from django.conf import settings
from django.contrib.auth import get_user_model
from django.urls import reverse_lazy
from jsmin import jsmin

from .forms import RegistrationForm, TwoFaForm
from ..utils import lazy_translate as __


UserModel = get_user_model()


class BaseView(TemplateView):
    login_required = False
    minify_response = True

    def dispatch(self, request, *args, **kwargs):
        response = super().dispatch(request, *args, **kwargs)
        if not self.minify_response:
            response.minify_response = False
        return response

    @classmethod
    def as_view(cls, *args, **kwargs):
        view = super().as_view(*args, **kwargs)
        return cls.login_required and login_required(view, login_url=reverse_lazy('login')) or view


class GUIView(BaseView):
    login_required = True
    template_name = "gui/gui.html"


class OfflineView(BaseView):
    login_required = False
    template_name = "gui/offline.html"


class ManifestView(BaseView):
    minify_response = False
    login_required = False
    template_name = "gui/manifest.json"
    content_type = 'application/json'


class JSMinTemplateResponse(TemplateResponse):
    @property
    def rendered_content(self):
        content = super().rendered_content
        return content.__class__(jsmin(content, quote_chars="'\"`"))


class SWView(BaseView):
    login_required = False
    minify_response = False
    content_type = 'text/javascript'
    template_name = "gui/service-worker.js"
    response_class = JSMinTemplateResponse


class Login(auth.LoginView):
    template_name = 'auth/login.html'
    redirect_authenticated_user = True

    def _is_need_tfa(self, request):
        return request.user.is_authenticated and getattr(request.user, 'need_twofa', False)

    def dispatch(self, request, *args, **kwargs):
        if self._is_need_tfa(request):
            self.redirect_authenticated_user = False
        return super(Login, self).dispatch(request, *args, **kwargs)

    def get_form_class(self):
        if self._is_need_tfa(self.request):
            return TwoFaForm
        return super().get_form_class()

    def get_template_names(self):
        if self._is_need_tfa(self.request):
            return ['auth/tfa.html']
        return super().get_template_names()

    def get_context_data(self, **kwargs):
        context = super(Login, self).get_context_data(**kwargs)
        context['wait_confirmation'] = self.request.GET.get('confirmation', '') == 'true'
        return context


class Logout(auth.LogoutView):
    next_page = reverse_lazy('login')


class Registration(FormView, BaseView):
    template_name = 'registration/user_registration.html'
    success_url = reverse_lazy('login')
    form_class = RegistrationForm

    def form_valid(self, form):
        form.request = self.request
        user = form.save()
        if settings.AUTHENTICATE_AFTER_REGISTRATION and user.id is not None:
            login(self.request, user)
        return super().form_valid(form)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if not settings.SEND_CONFIRMATION_EMAIL or 'uid' in context['form'].errors:
            return context
        context['uid'] = self.request.GET.get('uid', '')
        return context

    def get_success_url(self):
        url = super().get_success_url()
        if settings.SEND_CONFIRMATION_EMAIL and not self.request.POST.get('uid', False):
            url += '?confirmation=true'
        return url


class BaseAgreementsView(TemplateView):
    template_name = 'registration/base_agreements.html'
    title_message = __('Terms')
    path_in_settings: str

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['header_message'] = self.title_message
        path = Path(getattr(settings, self.path_in_settings, None))
        translated_file_path = path.with_name(path.name + f'.{self.request.language.code}')

        if translated_file_path.exists():
            path = translated_file_path

        try:
            context['terms'] = markdown(path.read_text(encoding='utf8'))
        except FileNotFoundError as e:
            raise Http404 from e

        return context


class ConsentToProcessingView(BaseAgreementsView):
    path_in_settings = 'CONSENT_TO_PROCESSING_PATH'
    title_message = __('personal data processing policy')


class TermsView(BaseAgreementsView):
    path_in_settings = 'AGREEMENT_TERMS_PATH'
    title_message = __('terms of agreement')



import time
import logging
import typing as _t

from django.db import connection
from django.apps import apps
from django.conf import settings
from django.http.request import HttpRequest
from django.http.response import HttpResponse
from django.utils import translation
from django.urls import resolve
from django.shortcuts import redirect

from .api.models import Language
from .utils import BaseVstObject


logger = logging.getLogger(settings.VST_PROJECT)


class QueryTimingLogger:
    __slots__ = ('queries_time',)

    def __init__(self):
        self.queries_time = 0

    def __call__(self, execute, sql, params, many, context):
        start = time.time()
        try:
            return execute(sql, params, many, context)
        finally:
            self.queries_time += time.time() - start


class BaseMiddleware(BaseVstObject):
    """
    Middleware base class for handling:

    * Incoming requests by :meth:`.BaseMiddleware.request_handler()`;
    * Outgoing response before any calling on server by :meth:`.BaseMiddleware.get_response_handler()`;
    * Outgoing responses by :meth:`.BaseMiddleware.handler()`.

    Middleware must be added to `MIDDLEWARE` list in settings.

    Example:
        .. sourcecode:: python

            from vstutils.middleware import BaseMiddleware
            from django.http import HttpResponse


            class CustomMiddleware(BaseMiddleware):
                def request_handler(self, request):
                    # Add header to request
                    request.headers['User-Agent'] = 'Mozilla/5.0'
                    return request

                def get_response_handler(self, request):
                    if not request.user.is_stuff:
                        # Return 403 HTTP status for non-stuff users.
                        # This request never gets in any view.
                        return HttpResponse(
                            "Access denied!",
                            content_type="text/plain",
                            status_code=403
                        )
                    return super().get_response_handler(request)

                def handler(self, request, response):
                    # Add header to response
                    response['Custom-Header'] = 'Some value'
                    return response

    """
    __slots__ = 'get_response', 'logger'

    logger: logging.Logger

    def __init__(self, get_response: _t.Callable):
        self.get_response = get_response
        self.logger = logger
        super().__init__()

    def get_setting(self, value: _t.Text):
        return self.get_django_settings(value)

    def handler(self, request: HttpRequest, response: HttpResponse) -> HttpResponse:  # nocv
        # pylint: disable=unused-argument

        """
        The response handler. Method to process responses.

        :param request: HTTP-request object.
        :type request: django.http.HttpRequest
        :param response: HTTP-response object which will be sended to client.
        :type response: django.http.HttpResponse
        :return: Handled response object.
        :rtype: django.http.HttpResponse
        """

        return response

    def request_handler(self, request: HttpRequest) -> HttpRequest:
        # pylint: disable=unused-argument

        """
        The request handler. Called before request is handled by a view.

        :param request: HTTP-request object which is wrapped from client request.
        :type request: django.http.HttpRequest
        :return: Handled request object.
        :rtype: django.http.HttpRequest
        """

        return request

    def get_response_handler(self, request: HttpRequest) -> HttpResponse:
        """
        Entrypoint for breaking or continuing request handling.
        This function must return `django.http.HttpResponse` object
        or result of parent class calling.

        :param request: HTTP-request object which is wrapped from client request.
        :type request: django.http.HttpRequest
        :rtype: django.http.HttpResponse
        """
        return self.get_response(request)

    def __call__(self, request: HttpRequest) -> HttpResponse:
        return self.handler(
            self.request_handler(request),
            self.get_response_handler(request)
        )


class TimezoneHeadersMiddleware(BaseMiddleware):
    __slots__ = ()

    def handler(self, request: HttpRequest, response: HttpResponse) -> HttpResponse:
        response['Server-Timezone'] = self.get_setting('TIME_ZONE')
        response['VSTutils-Version'] = self.get_setting('VSTUTILS_VERSION')
        return response


class ExecuteTimeHeadersMiddleware(BaseMiddleware):
    __slots__ = ()

    def __duration_handler(self, data):
        key, value = data
        if isinstance(value, (list, tuple, map, filter, _t.Generator)):
            value = ''.join((self.__duration_handler(('', v)) for v in value))
        elif isinstance(value, (int, float)):
            value = f';dur={float(value)}'
        elif isinstance(value, str) and value:
            if ' ' in value:
                value = f'"{value}"'
            value = f';desc={value}'
        elif not value:
            value = ''
        return f'{key}{value}'

    def _round_time(self, seconds: _t.Union[int, float]):
        return round(seconds * 1000, 2)

    def get_response_handler(self, request: HttpRequest) -> HttpResponse:
        start_time = time.time()
        get_response_handler = super().get_response_handler
        ql = QueryTimingLogger()

        if not getattr(request, 'is_bulk', False):
            with connection.execute_wrapper(ql):
                response = get_response_handler(request)
        else:
            response = get_response_handler(request)

        response_durations = getattr(response, 'timings', None)  # type: ignore
        total_time = self._round_time(time.time() - start_time)

        if getattr(request, 'is_bulk', False):
            response['Response-Time'] = str(total_time)
        else:
            if response_durations:
                response_durations = f', {", ".join(map(self.__duration_handler, response_durations.items()))}'
            else:
                response_durations = ""
            response_durations += f', db_execution_time;dur={self._round_time(ql.queries_time)}'
            response['Server-Timing'] = f'total;dur={total_time}{response_durations or ""}'
        return response


class LangMiddleware(BaseMiddleware):
    __slots__ = ()

    def get_lang_object(self, request: HttpRequest) -> _t.Tuple[Language, bool]:
        set_cookie = True
        if 'lang' in request.GET:
            code = request.GET['lang']
        elif settings.LANGUAGE_COOKIE_NAME in request.COOKIES:
            code = request.COOKIES[settings.LANGUAGE_COOKIE_NAME]
            set_cookie = False
        else:
            code = translation.get_language_from_request(request)  # type: ignore
        obj = Language.objects.filter(code=code).first()
        if obj is not None:
            return obj, set_cookie
        return Language.objects.get(code=settings.LANGUAGE_CODE), set_cookie  # nocv

    def get_response_handler(self, request: HttpRequest) -> HttpResponse:
        request.language, set_cookie = self.get_lang_object(request)  # type: ignore
        translation.activate(request.language.code)  # type: ignore
        request.LANGUAGE_CODE = translation.get_language()  # type: ignore
        response = super().get_response_handler(request)
        if set_cookie:
            response.set_cookie('lang', request.language.code, domain=settings.SESSION_COOKIE_DOMAIN)  # type: ignore
        if 'Content-Language' not in response:
            response['Content-Language'] = request.language.code  # type: ignore
        return response


class TwoFaMiddleware(BaseMiddleware):
    pass_names = (
        'login',
        'logout',
        'pwa_manifest',
        'service_worker',
        'offline_gui',
        'user_registration',
        'terms',
    )

    def request_handler(self, request: HttpRequest) -> HttpRequest:
        request.user.need_twofa = False  # type: ignore
        if request.user.is_authenticated:
            twofa = getattr(request.user, 'twofa', None)
            if twofa and not request.session.get('2fa', False):
                request.user.need_twofa = True  # type: ignore
        return request

    def check_url_name(self, request):
        url_name = resolve(request.path).url_name
        return any([
            url_name not in self.pass_names,
            url_name is not None and url_name.startswith('password_reset')
        ])

    def get_response_handler(self, request: HttpRequest) -> HttpResponse:
        if request.user.need_twofa and self.check_url_name(request):  # type: ignore
            return redirect('login')
        return super().get_response_handler(request)


class FrontendChangesNotifications(BaseMiddleware):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.notificator_class = apps.get_app_config('vstutils_api').module.notificator_class

    def get_response_handler(self, request: HttpRequest) -> HttpResponse:
        if settings.CENTRIFUGO_CLIENT_KWARGS:
            with self.notificator_class([]) as notificator:
                request.notificator = notificator  # type: ignore
                return super().get_response_handler(request)
        return super().get_response_handler(request)



import hashlib

import pyotp
from django.conf import settings
from django.db import models
from django.utils.module_loading import import_string as import_class
from django.utils.functional import cached_property

from ..models import BaseModel
from ..custom_model import ListModel, CharField


lib_names = []
for __attr_name in ['VST_PROJECT', 'VST_PROJECT_LIB_NAME']:
    __lib_name = getattr(settings, __attr_name)
    if __lib_name not in lib_names:
        lib_names.append(__lib_name)


class Language(ListModel):
    data = [
        {'code': code, 'name': name}
        for code, name in settings.LANGUAGES
    ]
    code = CharField(primary_key=True, max_length=5)
    name = CharField(max_length=128)

    @classmethod
    def get_etag_value(cls):
        hashable_str = '_'.join(c for c, _ in settings.LANGUAGES)
        if settings.ENABLE_CUSTOM_TRANSLATIONS:
            hashable_str += CustomTranslations.get_etag_value()
        return hashlib.md5(hashable_str.encode('utf-8')).hexdigest()

    def _get_translation_data(self, module_path_string, code, for_server=False):
        data = {}
        for dict_name in filter(bool, ['TRANSLATION'] + ['SERVER_TRANSLATION' if for_server else None]):
            try:
                translation_data = import_class(module_path_string + '.translations.' + code + f'.{dict_name}')
                if not isinstance(translation_data, dict):
                    continue
                data.update(translation_data.copy())
            except:
                pass
        return data

    def get_translations(self, for_server=False):
        code = self.code.replace('-', '_')
        translation_data = self._get_translation_data('vstutils', code, for_server)
        for attr_name in lib_names:
            translation_data.update(
                self._get_translation_data(attr_name, code, for_server)
            )
        if settings.ENABLE_CUSTOM_TRANSLATIONS:
            translation_data.update(
                CustomTranslations.objects.filter(code=code).values_list('original', 'translated')
            )
        return translation_data

    @cached_property
    def translations(self):
        return self.get_translations(for_server=False)

    @cached_property
    def server_translations(self):
        return self.get_translations(for_server=True)

    def translate(self, text):
        translated = self.server_translations.get(text, None)
        if translated is None:
            # place for additional translation methods
            return text
        return translated


class CustomTranslations(BaseModel):
    _cache_responses = True

    original = CharField(primary_key=True, max_length=380)
    translated = CharField(max_length=380)
    code = CharField(max_length=5)

    class Meta:
        default_related_name = 'custom_translations'
        indexes = [
            models.Index(fields=('original', 'translated', 'code'), name='%(app_label)s_translations')
        ]


class TwoFactor(BaseModel):
    user = models.OneToOneField(settings.AUTH_USER_MODEL, on_delete=models.CASCADE, primary_key=True)
    secret = models.CharField(max_length=256, blank=False, null=False)

    class Meta:
        default_related_name = 'twofa'
        indexes = [
            models.Index(fields=('user', 'secret'), name='%(app_label)s_tfa_fullidx')
        ]

    @property
    def enabled(self):
        return bool(self.secret)

    def verify(self, pin):
        if not pyotp.TOTP(self.secret).verify(pin):
            return self.recoverycode.filter(key=pin).delete()[0]
        return True


class RecoveryCode(BaseModel):
    id = models.AutoField(primary_key=True, max_length=100)
    key = models.CharField(max_length=256)
    tfa = models.ForeignKey(TwoFactor, on_delete=models.CASCADE)

    class Meta:
        default_related_name = 'recoverycode'
        indexes = [
            models.Index(fields=('id', 'key', 'tfa'), name='%(app_label)s_recov_fullidx')
        ]


class UserSettings(BaseModel):
    user = models.OneToOneField(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    value = models.JSONField(default=dict)

    class Meta:
        default_related_name = '_settings'



TRANSLATION = {
    'version': 'versions | version | versions',
    'app info': 'about',
    'on item': 'items | item | items',
    'file n selected': 'no files | 1 file selected | {n} files selected',
    'image n selected': 'no images | 1 image selected | {n} images selected',
    'imageValidationResolutionError': 'Must be between {min} and {max} px.',
    'instance': 'instance | instances',
    '_name': 'name',
}



from drf_yasg.views import get_schema_view
from rest_framework import permissions, versioning


class OpenApiView(get_schema_view()):  # type: ignore
    permission_classes = (permissions.AllowAny,)
    versioning_class = versioning.NamespaceVersioning



from typing import Any, Text, Dict, List, ClassVar, NoReturn, Tuple, Union, Type
from rest_framework import routers, views, request
from rest_framework.schemas import views as drf_views


UrlType = Text
ApiVersionType = Text
PrefixType = Text
ApiViewOptionKeyType = Text
ApiViewOptionValueType = Any
ApiViewOptionsType = Dict[ApiViewOptionKeyType, ApiViewOptionValueType]
ApiVersionStructureType = Dict[PrefixType, ApiViewOptionsType]
ApiStructureType = Dict[ApiVersionType, ApiVersionStructureType]
UrlListType = List[Tuple[PrefixType, Union[views.APIView, routers.DefaultRouter], Text]]


class _AbstractRouter(routers.DefaultRouter):
    custom_urls: UrlListType
    permission_classes: List
    create_schema: bool

    def register_view(self, prefix: PrefixType, view: views.APIView, name: Text = None) -> None:
        ...

    def unregister_view(self, prefix: PrefixType) -> None:
        ...

    def unregister(self, prefix: PrefixType) -> None:
        ...

    def generate(self, views_list: Dict) -> None:
        ...

    def _unreg(self, prefix: PrefixType, objects_list: UrlListType) -> UrlListType:
        ...

    def _get_custom_lists(self) -> List:
        ...

    def _get_views_custom_list(self, view_request: request.Request, registers: Any) -> Dict[PrefixType, UrlType]:
        ...

    def _get_api_root_dict(self) -> Dict:
        ...


class APIRouter(_AbstractRouter):
    root_view_name: Text
    router_version_name: Text

    def generate(self, views_list: ApiVersionStructureType) -> None:
        ...

    def _get_schema_view(self) -> drf_views.SchemaView:
        ...


class MainRouter(_AbstractRouter):
    routers: ClassVar[UrlListType]

    def __register_openapi(self) -> None:
        ...

    def register_router(self, prefix: PrefixType, router: APIRouter, name: Text = None) -> None:
        ...

    def unregister_router(self, prefix: PrefixType) -> None:
        ...

    def generate_routers(self, api: ApiStructureType) -> None:
        ...



import sys
from subprocess import check_call

from ._base import DockerCommand
from .web import get_celery_command


class Command(DockerCommand):
    help = "Run Celery for web-application"

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            'args',
            metavar='celery_arg=value', nargs='*',
            help='Args "name=value" for the celery app.',
        )
        parser.add_argument(
            '--nomigrate',
            action='store_false', dest='migrate', default=True,
            help="Do NOT run docker command for migration of databases.",
        )

    def handle(self, *args, **opts):
        super().handle(*args, **opts)

        if opts['migrate']:
            self.migrate(opts)

        cmd_args = dict(
            arg.split('=')
            for arg in args
        )

        cmd = f'{sys.executable} -m {get_celery_command(**cmd_args)}'
        self._print(f'Execute: {cmd}')
        sys.exit(check_call(
            cmd,
            stdout=sys.stdout,
            stderr=sys.stderr,
            stdin=sys.stdin,
            shell=True,  # nosec
        ))



# pylint: disable=unused-argument
import typing as _t

from django.conf import settings
from django.http import Http404
from rest_framework import (
    permissions as rest_permissions,
    throttling,
    request as drf_request,
    serializers as drf_serializers
)

from . import base, serializers, decorators as deco, responses, models
from .decorators import subaction
from ..utils import import_class


class LanguageSerializer(serializers.VSTSerializer):
    class Meta:
        model: _t.Type[models.Language] = models.Language
        fields: _t.Tuple = (
            'code',
            'name'
        )


class OneLanguageSerializer(serializers.VSTSerializer):
    translations: serializers.DataSerializer = serializers.DataSerializer(read_only=True)

    class Meta:
        model: _t.Type[models.Language] = models.Language
        fields: _t.Tuple = (
            'code',
            'name',
            'translations'
        )


class TranslationSerializer(serializers.VSTSerializer):
    """
    Serializer for API Endpoint that translates given phrase using dictionaries from vstutils/translations
    """
    original = drf_serializers.CharField()
    translated = drf_serializers.CharField(read_only=True)

    class Meta:
        model: _t.Type[models.Language] = models.Language
        fields: _t.Tuple = (
            'original',
            'translated'
        )

    def update(self, instance: models.Language, validated_data):
        original = validated_data['original']
        instance.original = original
        instance.translated = instance.translate(original)
        return instance


class SettingsViewSet(base.ListNonModelViewSet):
    """
    API endpoint that returns application settings.
    """
    base_name = "settings"

    def _get_localization_settings(self):
        return {
            "LANGUAGE_CODE": settings.LANGUAGE_CODE,
            "LANGUAGES": dict(settings.LANGUAGES),
            "USE_I18N": settings.USE_I18N,
            "USE_L10N": settings.USE_L10N,
            "TIME_ZONE": settings.TIME_ZONE,
            "USE_TZ": settings.USE_TZ
        }

    def _get_system_settings(self):
        return {
            "PY": settings.PY_VER,
            "VSTUTILS_VERSION": settings.VSTUTILS_VERSION,
            f"{settings.ENV_NAME}_VERSION": settings.PROJECT_VERSION
        }

    @deco.action(methods=['get'], detail=False)
    def localization(self, request: drf_request.Request):
        """
        Return localization settings.
        """
        return responses.HTTP_200_OK(self._get_localization_settings())

    @deco.action(methods=['get'], detail=False)
    def system(self, request: drf_request.Request):
        """
        Return system settings like interpreter or libs version.
        """
        return responses.HTTP_200_OK(self._get_system_settings())


class HealthThrottle(throttling.AnonRateThrottle):
    THROTTLE_RATES = {
        'anon': settings.HEALTH_THROTTLE_RATE
    }


class HealthView(base.ListNonModelViewSet):
    permission_classes = (rest_permissions.AllowAny,)
    authentication_classes = ()
    throttle_classes = (HealthThrottle,)
    health_backend = import_class(settings.HEALTH_BACKEND_CLASS)()

    def list(self, request, *args, **kwargs):
        return responses.HTTP_200_OK(*self.health_backend.get())


class LangViewSet(base.CachableHeadMixin, base.ReadOnlyModelViewSet):
    schema = None
    model: _t.Type[models.Language] = models.Language
    serializer_class: _t.Type[LanguageSerializer] = LanguageSerializer
    serializer_class_one: _t.Type[OneLanguageSerializer] = OneLanguageSerializer
    permission_classes: _t.Tuple[object] = (rest_permissions.IsAuthenticatedOrReadOnly,)

    def get_object(self):
        try:
            return super().get_object()
        except Http404:
            # Perform the lookup filtering.
            lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field

            assert lookup_url_kwarg in self.kwargs, (
                    'Expected view %s to be called with a URL keyword argument '
                    'named "%s". Fix your URL conf, or set the `.lookup_field` '
                    'attribute on the view correctly.' %
                    (self.__class__.__name__, lookup_url_kwarg)
            )

            obj_kwargs = {
                self.lookup_field: self.kwargs[lookup_url_kwarg],
                'name': self.kwargs[lookup_url_kwarg],
            }
            return self.model(**obj_kwargs)

    @subaction(
        methods=['post'],
        detail=True,
        serializer_class=TranslationSerializer,
    )
    def translate(self, request, pk=None):
        """
        detail action method for translating given phrases using dictionaries.
        """
        serializer = self.get_serializer(self.get_object(), data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return responses.HTTP_201_CREATED(serializer.data)



import typing as _t
import logging
from django.db.models import Model, QuerySet
from django.http.request import HttpRequest
from .utils import ObjectHandlers
from .ldap_utils import LDAP as __LDAP


_LDAP: _t.Type[__LDAP]
HAS_LDAP: bool
logger: logging.Logger


class ObjectDoesNotExist(Exception):
    """The requested object does not exist"""
    silent_variable_failure = True


class UserModel(Model):
    DoesNotExist: ObjectDoesNotExist


class BaseAuthBackend:
    def authenticate(self, request: HttpRequest, username=None, password=None):
        raise NotImplementedError  # nocv

    def patch_user_queryset(self, queryset: QuerySet) -> QuerySet:
        return queryset

    def user_can_authenticate(self, user: UserModel) -> bool:
        ...

    def get_user(self, user_id: int) -> _t.Union[UserModel, _t.NoReturn]:
        ...


class LDAP(__LDAP):
    ...


class LdapBackend(BaseAuthBackend):
    domain: _t.Union[_t.Text, _t.NoReturn]
    server: _t.Union[_t.Text, _t.NoReturn]

    def authenticate(self, request: HttpRequest, username: _t.Text = None, password: _t.Text = None) -> _t.Union[UserModel, _t.NoReturn]:
        ...


class AuthPluginsBackend(BaseAuthBackend):
    auth_handlers: _t.ClassVar[ObjectHandlers]
    auth_header: _t.ClassVar[_t.Text]

    def auth_with_plugin(self, plugin: _t.Text, request: HttpRequest, username: _t.Text, password: _t.Text) -> _t.Union[bool, _t.NoReturn]:
        return self.auth_handlers.get_object(plugin).authenticate(request, username, password)

    def authenticate(self, request: HttpRequest, username: _t.Text = None, password: _t.Text = None) -> _t.Union[bool, _t.NoReturn]:
        ...



from django.conf import settings
from django.apps import AppConfig

from ..utils import import_class


class VSTUtilsAppConfig(AppConfig):
    name = 'vstutils.api'
    label = 'vstutils_api'
    verbose_name = "vstutils"

    def ready(self):
        super().ready()
        self.module.autodiscover()
        self.module.notificator_class = import_class(
            settings.NOTIFICATOR_CLIENT_CLASS
        )



import traceback
from pathlib import Path
from logging import getLogger

from orjson import loads
from configparserc.tools import File
from django.conf import settings
from django.utils.functional import cached_property
from vstutils.utils import StaticFilesHandlers


logger = getLogger('vstutils')


class BaseStaticObjectHandler:
    def __init__(self, name, priority=254):
        self.name = name
        self.priority = priority

    def get_spa_static_list(self):
        return ()  # nocv

    @cached_property
    def spa_static_list(self):
        return self.get_spa_static_list()


class SPAStaticObjectHandler(BaseStaticObjectHandler):
    def get_spa_static_list(self):
        return tuple(settings.SPA_STATIC)


class WebpackJsonStaticObjectHandler(BaseStaticObjectHandler):
    def __init__(self, *args, **kwargs):
        self.json_file = kwargs.pop('json_file', str(Path(settings.VSTUTILS_DIR)/'static/bundle/output.json'))
        self.prefix = kwargs.pop('prefix', str(Path(self.json_file.split('/static/')[-1]).parent))
        self.entrypoint_name = kwargs.pop('entrypoint_name', kwargs.get('name'))
        super().__init__(*args, **kwargs)

    @cached_property
    def spa_static_list(self):
        return tuple(self.get_static_files_paths(self.get_static_files()))

    def get_static_files_paths(self, files_list):
        for obj in files_list:
            path = Path(obj['name'])
            priority = self.priority
            if path.stem != self.entrypoint_name:
                priority -= 1
            yield {
                'priority': priority,
                'type': path.suffix.replace('.', '') or 'js',
                'name': str(self.prefix / path),
                'source': self.name,
            }

    def get_static_files(self):
        with File(self.json_file) as fd:
            try:
                return loads(fd.read())['entrypoints'][self.entrypoint_name]['assets']
            except:  # nocv
                logger.exception(traceback.format_exc())


SPA_STATIC_FILES_PROVIDERS = StaticFilesHandlers('SPA_STATIC_FILES_PROVIDERS')



# pylint: disable=invalid-name
import os

from django.conf import settings
from django.urls.conf import include, re_path
from django.contrib import admin
from django.conf.urls.static import static
from django.views.static import serve
from django.contrib.staticfiles.urls import staticfiles_urlpatterns
from django.views.generic.base import RedirectView
from rest_framework import permissions

from .api.routers import MainRouter
from .utils import URLHandlers
from .api.views import HealthView


def get_valid_url(*args):
    return os.path.join('/', *map(lambda x: x.lstrip('^/'), filter(bool, args + ('/',))))


class AdminLoginLogoutRedirectView(RedirectView):
    query_string = True


# Main router for all APIs versions
router = MainRouter(perms=(permissions.IsAuthenticated,))
router.generate_routers(settings.API)
router.register_view('health', HealthView.as_view({'get': 'list'}), 'health')


admin.site.site_header = 'Admin panel'
admin.site.site_title = settings.VST_PROJECT
admin.site.index_title = f"{settings.VST_PROJECT.upper()} Settings Panel"
admin.site.site_url = "/"
admin.site.login = AdminLoginLogoutRedirectView.as_view(  # type: ignore
    url=get_valid_url(settings.ACCOUNT_URL, settings.LOGIN_URL)
)
admin.site.logout = AdminLoginLogoutRedirectView.as_view(  # type: ignore
    url=get_valid_url(settings.ACCOUNT_URL, settings.LOGOUT_URL)
)
doc_url = getattr(settings, 'DOC_URL', '/docs/')[1:]

urlpatterns = list(URLHandlers())

urlpatterns += [
    re_path(settings.ACCOUNT_URL, include(list(URLHandlers('ACCOUNT_URLS'))))
]

urlpatterns += [
    re_path(r'^admin/', admin.site.urls),
] if getattr(settings, 'ENABLE_ADMIN_PANEL', False) else []

urlpatterns += [re_path(rf'^{settings.API_URL}/', include(router.urls))]
if settings.STATIC_URL.startswith('/'):  # pragma: no branch
    urlpatterns += staticfiles_urlpatterns(settings.STATIC_URL)
if settings.MEDIA_URL.startswith('/') and settings.MEDIA_ROOT:
    urlpatterns += static(settings.MEDIA_URL, view=serve, document_root=settings.MEDIA_ROOT)  # nocv
if 'docs' in settings.INSTALLED_APPS:  # nocv
    urlpatterns += [re_path(rf'^{doc_url}', include(('docs.urls', settings.VST_PROJECT), namespace='docs'))]



# Generated by Django 2.2.2 on 2019-12-09 01:49

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [  # type: ignore
    ]

    operations = [
        migrations.CreateModel(
            name='Language',
            fields=[
                ('code', models.CharField(max_length=5, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=128)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import functools
import operator

import ormsgpack
from rest_framework.exceptions import ParseError
from rest_framework.parsers import BaseParser
from rest_framework.settings import api_settings

from .renderers import MsgpackRenderer


class MsgpackParser(BaseParser):
    """
    Parses MessagePack-serialized data.
    """

    media_type = 'application/msgpack'
    renderer_class = MsgpackRenderer
    options = functools.reduce(
        operator.or_,
        api_settings.user_settings.get("MSGPACK_PARSER_OPTIONS", ()),
        None,
    )

    def parse(self, stream, media_type=None, parser_context=None):
        try:
            return ormsgpack.unpackb(stream.read(), option=self.options)
        except Exception as exc:  # nocv
            raise ParseError(f'Msgpack parse error - {exc}') from exc



# Generated by Django 3.2 on 2021-04-09 04:07

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('vstutils_api', '0003_tfa_indexes'),
    ]

    operations = [
        migrations.CreateModel(
            name='UserSettings',
            fields=[
                ('id', models.AutoField(max_length=20, primary_key=True, serialize=False)),
                ('hidden', models.BooleanField(default=False)),
                ('value', models.JSONField(default=dict)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddIndex(
            model_name='recoverycode',
            index=models.Index(fields=['id', 'key', 'tfa'], name='vstutils_api_recov_fullidx'),
        ),
        migrations.AddIndex(
            model_name='twofactor',
            index=models.Index(fields=['user', 'secret'], name='vstutils_api_tfa_fullidx'),
        ),
        migrations.AddField(
            model_name='usersettings',
            name='user',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='_settings', to=settings.AUTH_USER_MODEL),
        ),
    ]



import copy
import heapq

# pylint: disable=import-error,signature-differs
from django_celery_beat.schedulers import DatabaseScheduler
from celery.beat import event_t
from celery.utils.log import get_logger

from .utils import Lock

logger = get_logger(__name__)


class SingletonDatabaseScheduler(DatabaseScheduler):
    scheduler_lock = None
    discarding_overdue_tasks = True

    def _discard_overdue_tasks(self, heappop, heappush):
        """
        Discards overdued tasks on first `tick` function call.
        It emulates executing overdued task,
        uses standart `populate_heap` function
        where overdued tasks have higher priority and
        therefore they are first in heap.

        :return: preferred delay in seconds for next tick call.
        :rtype: ``int``
        """
        if (self._heap is None or
                not self.schedules_equal(self.old_schedulers, self.schedule)):
            self.old_schedulers = copy.copy(self.schedule)
            self.populate_heap()

        events_heap = self._heap

        if not events_heap:
            return self.max_interval

        for _i in range(len(events_heap)):
            event = events_heap[0]
            entry = event[2]
            is_due, next_time_to_run = self.is_due(entry)
            if not is_due:
                break
            verify = heappop(events_heap)
            if verify is event:
                next_entry = self.reserve(entry)
                self._tasks_since_sync += 1
                if self.should_sync():
                    self._do_sync()
                logger.info(f'{entry.task} task discarded!')
                heappush(events_heap, event_t(self._when(next_entry, next_time_to_run),
                                              event[1], next_entry))
        return 0

    # pylint: disable=redefined-outer-name
    def tick(self, heappop=heapq.heappop, heappush=heapq.heappush, *args, **kwargs):
        """
        Run a tick - one iteration of the scheduler.
        Discards overdued tasks on first call.
        Executes one due task per call.

        :return: preferred delay in seconds for next tick call.
        :rtype: ``int``
        """
        if self.discarding_overdue_tasks:
            self.discarding_overdue_tasks = False
            return self._discard_overdue_tasks(heappop, heappush)

        if self.scheduler_lock is None:
            try:
                self.scheduler_lock = Lock(Lock.SCHEDULER, timeout=120.0)
            except Lock.AcquireLockException:
                return 60.0
        self.scheduler_lock.prolong()
        return super().tick(event_t, min, heappop, heappush)

    def close(self):
        if self.scheduler_lock is not None:
            self.scheduler_lock.release()

        return super().close()



# pylint: disable=abstract-method,unused-import,undefined-variable
import json
import os
import logging
import tempfile
import sys
import time
import traceback
from subprocess import check_call
from collections import OrderedDict

from django.conf import settings
from django.core.management.base import (
    BaseCommand as _BaseCommand,
    CommandError as CmdError,
)
from configparserc.config import ConfigParserC, Section, BoolType
from vstutils.utils import Lock


tmp = tempfile.tempdir
logger = logging.getLogger(settings.VST_PROJECT)
logger_lib = logging.getLogger(settings.VST_PROJECT_LIB)


class BaseCommand(_BaseCommand):
    interactive = False
    requires_system_checks = False
    keep_base_opts = False
    help = "Service command for web-application"

    class CommandError(CmdError):
        pass

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.fromfile_prefix_chars = '@'
        parser.add_argument(
            '-l', '--log-level',
            action='store',
            dest='log-level',
            default=False,
            type=str,
            help='Set logs level [debug|warning|error|critical]')
        if self.interactive:
            parser.add_argument(
                '--noinput', '--no-input',
                action='store_false', dest='interactive', default=True,
                help="Do NOT prompt the user for input of any kind.",
            )

    def _print(self, msg, style=None):
        style = style or 'HTTP_INFO'
        style = getattr(self.style, style, str)
        self.stdout.write(style(msg))

    def _settings(self, value, default=None):
        return getattr(settings, value, default)

    def _get_versions(self):
        versions = OrderedDict(django=super().get_version())
        versions[self._settings('VST_PROJECT')] = self._settings('PROJECT_VERSION')
        if self._settings('VST_PROJECT') != self._settings('VST_PROJECT_LIB'):
            versions[self._settings('VST_PROJECT_LIB')] = self._settings(
                'PROJECT_LIB_VERSION'
            )
        if self._settings('VST_PROJECT') != "vstutils":
            versions['vstutils'] = self._settings('VSTUTILS_VERSION')
        return versions

    def get_version(self):
        return ' '.join([
            f'{k.title()}={v}' for k, v in self._get_versions().items()
        ])

    def ask_user(self, message, default=None):
        # pylint: disable=import-error
        if getattr(self, 'interactive_mode', False):
            return input(message) or default
        return default

    def ask_user_bool(self, message, default=True):
        reply = self.ask_user(message, 'yes' if default else 'no').lower()
        if reply in ['y', 'yes']:
            return True
        elif reply in ['n', 'no']:
            return False

    def handle(self, *args, **options):
        # pylint: disable=invalid-name
        self.interactive_mode = options.pop('interactive', False)
        LOG_LEVEL = settings.LOG_LEVEL
        if options.get('log-level', False):
            LOG_LEVEL = options.pop('log-level', LOG_LEVEL)
        logger.setLevel(LOG_LEVEL.upper())
        logger_lib.setLevel(LOG_LEVEL.upper())
        self.LOG_LEVEL = LOG_LEVEL.upper()
        os.environ.setdefault('DJANGO_LOG_LEVEL', self.LOG_LEVEL)


class DockerCommand(BaseCommand):
    interactive = True
    with_migration = True

    def add_arguments(self, parser):
        super().add_arguments(parser)
        if self.with_migration:
            parser.add_argument(
                '--migrate-attempts', '-a',
                default=60,
                dest='attempts', help='The number of attempts to migrate.',
            )
            parser.add_argument(
                '--migrate-attempts-sleep-time', '-t',
                default=1,
                dest='attempts_timeout', help='The number of attempts to migrate.',
            )

    def handle(self, *args, **options):
        super().handle(*args, **options)
        self.prefix = self._settings('VST_PROJECT_LIB', 'vstutils').upper()
        self.project_name = self._settings('VST_PROJECT', 'vstutils')

        logger.debug(f'Prefix={self.prefix} | Project={self.project_name}')

        self.env = os.environ.copy()
        config = self.prepare_config()
        self.env[self._settings('CONFIG_ENV_DATA_NAME')] = config.generate_config_string()
        default_envs = {
            'UWSGI_PROCESSES': 'UWSGI_WORKERS',
            'UWSGI_THREADS': 'UWSGI_THREADS'
        }
        for key in default_envs:  # pylint: disable=consider-using-dict-items
            value = os.environ.get(f"{self.prefix}_{key}", '')
            if value:
                self.env[default_envs[key]] = value  # nocv

        if config['main']['debug'] or self._settings('TESTS_RUN', False):
            logger.debug(f'Env:\n{json.dumps(self.env, indent=4)}')
            logger.debug(f'Config:\n{self.env[self._settings("CONFIG_ENV_DATA_NAME")]}')

    @property
    def databases_to_migrate(self):
        allowed_databases = ('default',) + tuple(set(getattr(settings, 'DOCKER_DATABASES_TO_MIGRATE', ())))
        for db in settings.DATABASES.keys():
            if db in allowed_databases:
                yield db

    def migrate(self, options, *args):
        success = False
        error = 'Unknown error.'
        for i in range(options.get('attempts', 60)):
            try:
                with Lock(
                        self._settings('DOCKERRUN_MIGRATE_LOCK_ID'),
                        1,
                        self._settings('DOCKERRUN_MIGRATE_LOCK_TIMEOUT'),
                        'Migration process still locked by another application process.',
                        int(os.getenv("DOCKER_MIGRATE_LOCK_KEY_TIMEOUT", '0')) or None
                ) as lock:  # noqa: F841, pylint:disable=unused-variable
                    logger.info(f'Migration locked by key: `{lock.id}`')
                    for db_name in self.databases_to_migrate:
                        logger.info(f'Migrating db "{db_name}".')
                        check_call(
                            [sys.executable, '-m', self.project_name, 'migrate', '--database', db_name, *args],
                            env=self.env,
                            bufsize=0,
                            universal_newlines=True,
                        )
                    logger.info(f'Unlocking migration by key: `{lock.id}`')
            except:
                error = traceback.format_exc()
                logger.debug(f'Migration attempt {i} failed: {sys.exc_info()[1]}')
                self._print(f"Retry #{i}...", 'WARNING')
                time.sleep(options.get('attempts_timeout', 1))
            else:
                success = True
                break
        return success, error

    def prepare_section_main(self, config):
        config['main'] = {
            'debug': os.getenv(f'{self.prefix}_DEBUG', 'false'),
            'log_level': self.log_level,
            'timezone': os.getenv(f'{self.prefix}_TIMEZONE', 'UTC'),
            'enable_admin_panel': os.getenv(f'{self.prefix}_ENABLE_ADMIN_PANEL', 'false'),
            'first_day_of_week': os.getenv(
                f'{self.prefix}_FIRST_DAY_OF_WEEK',
                self._settings('FIRST_DAY_OF_WEEK')
            )
        }
        # ldap-server, ldap-default-domain if exist
        ldap_server = os.getenv(f'{self.prefix}_LDAP_CONNECTION', None)
        if ldap_server:  # nocv
            config['main']['ldap_server'] = ldap_server
        ldap_default_domain = os.getenv(f'{self.prefix}_LDAP_DOMAIN', None)
        if ldap_default_domain:  # nocv
            config['main']['ldap-default-domain'] = ldap_default_domain

    def prepare_section_db(self, config):
        # SQLite prepearing
        sqlite_default_dir = os.environ.get(f'{self.prefix}_SQLITE_DIR', '/')
        if sqlite_default_dir != '/' and not os.path.exists(sqlite_default_dir):  # nocv
            os.makedirs(sqlite_default_dir)
        if sqlite_default_dir[-1] != '/':  # nocv
            sqlite_default_dir += '/'
        sqlite_default_name = os.environ.get(f'{self.prefix}_SQLITE_DBNAME', 'db.sqlite3')
        sqlite_db_path = f'{sqlite_default_dir}/{sqlite_default_name}'

        if os.getenv(f'{self.prefix}_DB_HOST') is not None:
            try:
                pm_type = os.getenv(f'{self.prefix}_DB_TYPE', 'mysql')

                default_port = ''
                if pm_type == 'mysql':
                    default_port = '3306'
                elif pm_type == 'postgresql':  # nocv
                    default_port = '5432'

                config['database'] = {
                    'engine': f'django.db.backends.{pm_type}',
                    'name': os.environ[f'{self.prefix}_DB_NAME'],
                    'user': os.environ[f'{self.prefix}_DB_USER'],
                    'password': os.environ[f'{self.prefix}_DB_PASSWORD'],
                    'host': os.environ[f'{self.prefix}_DB_HOST'],
                    'port': os.getenv(f'{self.prefix}_DB_PORT', default_port),
                }
                config['database.options'] = {
                    'connect_timeout': os.getenv(f'{self.prefix}_DB_CONNECT_TIMEOUT', '20'),
                }
                if pm_type == 'mysql':
                    config['database.options']['init_command'] = os.getenv('DB_INIT_CMD', '')
            except KeyError as err:  # nocv
                raise Exception('Not enough variables for connect to  SQL server.') from err
        else:  # nocv
            config['database'] = {
                'engine': 'django.db.backends.sqlite3',
                'name': sqlite_db_path
            }

    def prepare_section_cache(self, config):
        cache_loc = os.getenv('CACHE_LOCATION', f'{tmp}/{self.prefix}_django_cache')
        cache_type = os.getenv(f'{self.prefix}_CACHE_TYPE', 'file')
        if cache_type == 'file':
            cache_engine = 'django.core.cache.backends.filebased.FileBasedCache'
        elif cache_type == 'memcache':  # nocv
            cache_engine = 'django.core.cache.backends.memcached.MemcachedCache'
        elif cache_type == 'redis':  # nocv
            cache_engine = 'django_redis.cache.RedisCache'
        else:  # nocv
            raise Exception(f'Unknown cache type `{cache_type}`.')

        config['cache'] = config['locks'] = {
            'backend': cache_engine,
            'location': cache_loc
        }

    def prepare_section_rpc(self, config):
        rpc_connection = os.getenv('RPC_ENGINE', None)
        config['rpc'] = {
            'heartbeat': os.getenv('RPC_HEARTBEAT', '5'),
            'concurrency': os.getenv('RPC_CONCURRENCY', '4'),
            'clone_retry_count': os.getenv('RPC_CLONE_RETRY_COUNT', '3')
        }

        if rpc_connection:  # nocv
            config['rpc']['connection'] = rpc_connection
        config['rpc']['enable_worker'] = 'false'

        # Set worker settings
        if os.environ.get('WORKER', '') == 'ENABLE':  # nocv
            config['rpc']['enable_worker'] = 'true'
            config['worker'] = {
                'loglevel': self.log_level,
                'pidfile': f'{tmp}/{self.prefix.lower()}_worker.pid',
                'beat': os.getenv(f'{self.prefix}_SCHEDULER_ENABLE', 'true')
            }

    def prepare_section_web(self, config):
        config['web'] = {
            'session_timeout': os.getenv(f'{self.prefix}_SESSION_TIMEOUT', '2w'),
            'rest_page_limit': os.getenv(f'{self.prefix}_WEB_REST_PAGE_LIMIT', '100'),
            'enable_gravatar': os.getenv(f'{self.prefix}_WEB_GRAVATAR', 'true'),
            'request_max_size': os.getenv(
                f'{self.prefix}_REQUEST_MAX_SIZE',
                self._settings('DATA_UPLOAD_MAX_MEMORY_SIZE')
            ),
            'x_frame_options': os.getenv(
                f'{self.prefix}_X_FRAME_OPTIONS',
                self._settings('X_FRAME_OPTIONS')
            ),
            'use_x_forwarded_host': os.getenv(
                f'{self.prefix}_USE_X_FORWARDED_HOST',
                self._settings('USE_X_FORWARDED_HOST')
            ),
            'use_x_forwarded_port': os.getenv(
                f'{self.prefix}_USE_X_FORWARDED_PORT',
                self._settings('USE_X_FORWARDED_PORT')
            ),
            'password_reset_timeout_days': os.getenv(
                f'{self.prefix}_PASSWORD_RESET_TIMEOUT_DAYS',
                self._settings('PASSWORD_RESET_TIMEOUT_DAYS')
            ),
            'secure_browser_xss_filter': os.getenv(
                f'{self.prefix}_SECURE_BROWSER_XSS_FILTER',
                self._settings('SECURE_BROWSER_XSS_FILTER')
            ),
            'secure_content_type_nosniff': os.getenv(
                f'{self.prefix}_SECURE_CONTENT_TYPE_NOSNIFF',
                self._settings('SECURE_CONTENT_TYPE_NOSNIFF')
            ),
            'secure_hsts_include_subdomains': os.getenv(
                f'{self.prefix}_SECURE_HSTS_INCLUDE_SUBDOMAINS',
                self._settings('SECURE_HSTS_INCLUDE_SUBDOMAINS')
            ),
            'secure_hsts_preload': os.getenv(
                f'{self.prefix}_SECURE_HSTS_PRELOAD',
                self._settings('SECURE_HSTS_PRELOAD')
            ),
            'secure_hsts_seconds': os.getenv(
                f'{self.prefix}_SECURE_HSTS_SECONDS',
                self._settings('SECURE_HSTS_SECONDS')
            ),
            'health_throttle_rate': os.getenv(
                f'{self.prefix}_HEALTH_THROTTLE_RATE',
                self._settings('HEALTH_THROTTLE_RATE').replace('/minute', '')
            ),
        }

    def prepare_section_uwsgi(self, config):
        config['uwsgi'] = {
            'thread-stacksize': os.getenv(f'{self.prefix}_UWSGI_THREADSTACK', '40960'),
            'max-requests': os.getenv(f'{self.prefix}_UWSGI_MAXREQUESTS', '50000'),
            'limit-as': os.getenv(f'{self.prefix}_UWSGI_LIMITS', '512'),
            'pidfile': os.getenv(f'{self.prefix}_UWSGI_PIDFILE', '/run/web.pid'),
            'daemon': 'false'
        }
        current_addr, current_port = self._settings('WEB_ADDRPORT').split(',')[0].split(':')
        config['uwsgi']['addrport'] = (
            f"{os.getenv(f'{self.prefix}_WEB_HOST', current_addr)}:"
            f"{os.getenv(f'{self.prefix}_WEB_PORT', current_port)}"
        )
        harakiri = os.getenv(f'{self.prefix}_UWSGI_HARAKIRI', '')
        if harakiri:
            config['uwsgi']['harakiri'] = harakiri
        vacuum = os.getenv(f'{self.prefix}_UWSGI_VACUUM', '')
        if vacuum:
            config['uwsgi']['vacuum'] = vacuum

    def prepare_section_smtp(self, config):
        mail_parameters = ['port', 'user', 'password', 'tls', 'ssl', 'from_address']
        mail_settings = {'host': os.environ.get(f'{self.prefix}_MAIL_HOST')}
        if mail_settings['host']:
            for param in mail_parameters:
                value = os.environ.get(f'{self.prefix}_MAIL_{param.upper()}')
                if value:
                    mail_settings[param] = value
            config['mail'] = mail_settings

    def prepare_config(self):
        # pylint: disable=too-many-locals,too-many-branches,too-many-statements
        prefix = self.prefix

        default_sections = (
            'main',
            'db',
            'cache',
            'rpc',
            'web',
            'uwsgi',
            'smtp',
        )
        section_handlers_prefix = 'prepare_section_'
        sections = default_sections + tuple(map(
            lambda y: y.replace(section_handlers_prefix, ''),
            filter(
                lambda x: x.startswith(section_handlers_prefix) and x not in default_sections,
                dir(self)
            )
        ))

        class DockerSection(Section):
            types_map = {
                f'override_{s}': BoolType()
                for s in sections
            }

        # Start configuring config file
        self.config = ConfigParserC(
            format_kwargs=self._settings('KWARGS', {}),
            section_defaults={
                'docker': {
                    s: os.getenv(f'{prefix}_DOCKER_{s.upper()}', 'True')
                    for s in DockerSection.types_map
                }
            },
            section_overload={
                'docker': DockerSection
            },
            format_exclude_sections=('uwsgi',)
        )
        config = self.config
        config.parse_files(self._settings('CONFIG_FILES'))

        # Set log level
        self.log_level = os.getenv(f'{prefix}_LOG_LEVEL', 'WARNING')

        for section_name in sections:
            if config['docker'][f'override_{section_name}']:
                getattr(self, f'prepare_section_{section_name}')(config)

        # Set secret key
        os.environ.setdefault('SECRET_KEY', 'DISABLE')
        if os.environ['SECRET_KEY'] != 'DISABLE':  # nocv
            with open(f'/etc/{prefix.lower()}/secret', 'w', encoding='utf-8') as secretfile:
                secretfile.write(os.environ['SECRET_KEY'])

        return config



from copy import deepcopy
from typing import Dict, Type, Text, Any
from collections import OrderedDict

from django.http import FileResponse
from django.db import models
from django.utils.functional import cached_property
from drf_yasg.inspectors.base import FieldInspector, NotHandled
from drf_yasg.inspectors.field import ReferencingSerializerInspector, decimal_field_type
from drf_yasg import openapi
from drf_yasg.inspectors.query import CoreAPICompatInspector, force_real_str, coreschema  # type: ignore
from rest_framework.fields import Field, JSONField, DecimalField, empty

from .. import fields, serializers, validators
from ...models.base import get_first_match_name


# Extra types

# Extra formats
FORMAT_FILE = openapi.TYPE_FILE
FORMAT_SECRET_FILE = 'secretfile'  # nosec
FORMAT_BIN_FILE = 'binfile'
FORMAT_CSV_FILE = 'csvfile'
FORMAT_NAMED_BIN_FILE = 'namedbinfile'
FORMAT_NAMED_BIN_IMAGE = 'namedbinimage'
FORMAT_AUTOCOMPLETE = 'autocomplete'
FORMAT_FK_AUTOCOMPLETE = 'fk_autocomplete'
FORMAT_MULTISELECT = 'multiselect'
FORMAT_HTML = 'html'
FORMAT_JSON = 'json'
FORMAT_TEXTAREA = 'textarea'
FORMAT_DYN = 'dynamic'
FORMAT_DYN_FK = 'dynamic_fk'
FORMAT_FK = 'fk'
FORMAT_UPTIME = 'uptime'
FORMAT_RELATED_LIST = 'related_list'
FORMAT_RATING = 'rating'
FORMAT_PHONE = 'phone'
FORMAT_MASKED = 'masked'
FORMAT_DEEP_FK = 'deep_fk'
FORMAT_DECIMAL = 'decimal'
FORMAT_WYSIWYG = 'wysiwyg'

X_OPTIONS = 'x-options'


# Base types
basic_type_info: Dict[Type[Field], Dict[Text, Any]] = OrderedDict()
basic_type_info[fields.FileInStringField] = {
    'type': openapi.TYPE_STRING,
    'format': FORMAT_FILE
}
basic_type_info[fields.SecretFileInString] = {
    'type': openapi.TYPE_STRING,
    'format': FORMAT_SECRET_FILE
}
basic_type_info[fields.BinFileInStringField] = {
    'type': openapi.TYPE_STRING,
    'format': FORMAT_BIN_FILE
}

basic_type_info[fields.NamedBinaryFileInJsonField] = {
    'type': openapi.TYPE_OBJECT,
    'x-format': FORMAT_NAMED_BIN_FILE,
    'properties': {
        k: openapi.Schema(
            type=openapi.TYPE_STRING,
            default=v,
        )
        for k, v in fields.DEFAULT_NAMED_FILE_DATA.items()
    }
}
basic_type_info[fields.MultipleNamedBinaryFileInJsonField] = {
    'type': openapi.TYPE_ARRAY,
    'items': openapi.Items(
        **basic_type_info[fields.NamedBinaryFileInJsonField]
    )
}
basic_type_info[fields.HtmlField] = {
    'type': openapi.TYPE_STRING,
    'format': FORMAT_HTML
}
basic_type_info[serializers.JsonObjectSerializer] = {
    'type': openapi.TYPE_OBJECT,
    'format': FORMAT_JSON
}
basic_type_info[JSONField] = {
    'type': openapi.TYPE_STRING,
    'format': FORMAT_JSON
}
basic_type_info[fields.TextareaField] = {
    'type': openapi.TYPE_STRING,
    'format': FORMAT_TEXTAREA
}
basic_type_info[fields.UptimeField] = {
    'type': openapi.TYPE_INTEGER,
    'format': FORMAT_UPTIME
}
basic_type_info[fields.RedirectIntegerField] = {
    'type': openapi.TYPE_INTEGER
}
basic_type_info[fields.RedirectCharField] = {
    'type': openapi.TYPE_STRING
}
basic_type_info[fields.PasswordField] = {
    'type': openapi.TYPE_STRING,
    'format': openapi.FORMAT_PASSWORD
}
basic_type_info[fields.PhoneField] = {
    'type': openapi.TYPE_STRING,
    'format': FORMAT_PHONE,
}
basic_type_info[fields.WYSIWYGField] = {
    'type': openapi.TYPE_STRING,
    'format': FORMAT_WYSIWYG,
}


def field_have_redirect(field, **kwargs):
    if not getattr(field, 'redirect', False):
        return kwargs

    if kwargs.get(X_OPTIONS, None) is None:
        kwargs[X_OPTIONS] = {}
    operation_name = getattr(field, 'operation_name', None)
    if operation_name is None:
        operation_name = field.field_name
        splitted_operation_name = operation_name.split('_')
        if len(tuple(filter(bool, splitted_operation_name))) > 1:
            operation_name = '_'.join(splitted_operation_name[:-1])
    kwargs[X_OPTIONS]['redirect'] = {
        'operation_name': operation_name,
        'depend_field': getattr(field, 'depend_field', None),
        'concat_field_name': getattr(field, 'concat_field_name', False),
    }

    return kwargs


def field_extra_handler(field, **kwargs):
    kwargs = field_have_redirect(field, **kwargs)
    if kwargs['type'] in (openapi.TYPE_ARRAY, openapi.TYPE_OBJECT):
        kwargs['title'] = force_real_str(field.label) if field.label else None
    return kwargs


def _get_handled_props(group_mapping):
    for group_name, props in (group_mapping or {}).items():
        if group_name == '':
            continue
        yield from props


class VSTFieldInspector(FieldInspector):
    def field_to_swagger_object(self, field, swagger_object_type, use_references, **kw):
        # pylint: disable=invalid-name,unused-variable
        type_info = basic_type_info.get(field.__class__, None)
        if type_info is None:
            return NotHandled

        SwaggerType, ChildSwaggerType = self._get_partial_types(
            field, swagger_object_type, use_references, **kw
        )
        return SwaggerType(**field_extra_handler(field, **deepcopy(type_info)))


class AutoCompletionFieldInspector(FieldInspector):
    def field_to_swagger_object(self, field, swagger_object_type, use_references, **kw):
        # pylint: disable=invalid-name
        if not isinstance(field, fields.AutoCompletionField):
            return NotHandled

        SwaggerType, _ = self._get_partial_types(
            field, swagger_object_type, use_references, **kw
        )
        kwargs = {
            'type': openapi.TYPE_STRING,
            'format': FORMAT_AUTOCOMPLETE
        }
        if isinstance(field.autocomplete, (list, tuple)):
            kwargs['enum'] = list(field.autocomplete)
        else:
            kwargs['format'] = FORMAT_FK_AUTOCOMPLETE
            kwargs[X_OPTIONS] = {
                'model': openapi.SchemaRef(
                    self.components.with_scope(openapi.SCHEMA_DEFINITIONS),
                    field.autocomplete, ignore_unresolved=True
                ),
                'value_field': field.autocomplete_property,
                'view_field': field.autocomplete_represent,
                'usePrefetch': field.use_prefetch,
            }

        return SwaggerType(**field_extra_handler(field, **kwargs))


class DynamicJsonTypeFieldInspector(FieldInspector):
    def field_to_swagger_object(self, field, swagger_object_type, use_references, **kw):
        # pylint: disable=unused-variable,invalid-name
        if not isinstance(field, fields.DynamicJsonTypeField):
            return NotHandled

        SwaggerType, ChildSwaggerType = self._get_partial_types(
            field, swagger_object_type, use_references, **kw
        )
        options = {'field': field.field}

        if isinstance(field, fields.DependFromFkField):
            field_format = FORMAT_DYN_FK
            options['field_attribute'] = field.field_attribute

        else:
            field_format = FORMAT_DYN
            options['choices'] = field.choices
            options['types'] = {}
            for name, field_type in field.types.items():
                if isinstance(field_type, Field):
                    field_type = self.probe_field_inspectors(field_type, swagger_object_type, False)
                options['types'][name] = field_type

        kwargs = {
            'type': openapi.TYPE_STRING,
            'format': field_format,
            X_OPTIONS: options
        }

        return SwaggerType(**field_extra_handler(field, **kwargs))


class FkFieldInspector(FieldInspector):
    def field_to_swagger_object(self, field, swagger_object_type, use_references, **kw):
        # pylint: disable=unused-variable,invalid-name
        if not isinstance(field, fields.FkField):
            return NotHandled

        SwaggerType, ChildSwaggerType = self._get_partial_types(
            field, swagger_object_type, use_references, **kw
        )

        field_format = FORMAT_FK
        options = {
            'model': openapi.SchemaRef(
                self.components.with_scope(openapi.SCHEMA_DEFINITIONS),
                field.select_model, ignore_unresolved=True
            ),
            'value_field': field.autocomplete_property,
            'view_field': field.autocomplete_represent,
            'usePrefetch': field.use_prefetch,
            'makeLink': field.make_link,
            'dependence': field.dependence,
            'filters': field.filters,
        }

        if isinstance(field, fields.DeepFkField):
            field_format = FORMAT_DEEP_FK
            options = {
                **options,
                'only_last_child': field.only_last_child,
                'parent_field_name': field.parent_field_name,
            }
            del options['dependence']

        kwargs = {
            'type': openapi.TYPE_INTEGER,
            'format': field_format,
            X_OPTIONS: options,
        }

        return SwaggerType(**field_extra_handler(field, **kwargs))


class CommaMultiSelectFieldInspector(FieldInspector):
    def field_to_swagger_object(self, field, swagger_object_type, use_references, **kw):
        # pylint: disable=unused-variable,invalid-name
        if not isinstance(field, fields.CommaMultiSelect):
            return NotHandled

        SwaggerType, ChildSwaggerType = self._get_partial_types(
            field, swagger_object_type, use_references, **kw
        )
        kwargs = {
            'type': 'array',
            'x-collectionFormat': 'csv',
            X_OPTIONS: {
                'viewSeparator': field.select_separator,
            },
            'items': {
                "type": openapi.TYPE_INTEGER,
                "format": FORMAT_FK,
                X_OPTIONS: {
                    'model': openapi.SchemaRef(
                        self.components.with_scope(openapi.SCHEMA_DEFINITIONS),
                        field.select_model,
                        ignore_unresolved=True,
                    ),
                    "value_field": field.select_property,
                    "view_field": field.select_represent,
                    'usePrefetch': field.use_prefetch,
                    'makeLink': field.make_link,
                    'dependence': field.dependence,
                },
            }
        }

        return SwaggerType(**field_extra_handler(field, **kwargs))


class RelatedListFieldInspector(FieldInspector):
    def field_to_swagger_object(self, field, swagger_object_type, use_references, **kw):
        # pylint: disable=unused-variable,invalid-name
        if not isinstance(field, fields.RelatedListField):
            return NotHandled

        SwaggerType, ChildSwaggerType = self._get_partial_types(
            field, swagger_object_type, use_references, **kw
        )
        serializer_schema = VSTReferencingSerializerInspector(
            self.view,
            self.path,
            self.method,
            self.components,
            self.request,
            self.field_inspectors
        ).field_to_swagger_object(field.serializer_class(), swagger_object_type, False, **kw)

        kwargs = {
            'type': openapi.TYPE_ARRAY,
            'x-format': field.view_type,
            'items': serializer_schema,
        }

        return SwaggerType(**field_extra_handler(field, **kwargs))


class RatingFieldInspector(FieldInspector):
    def field_to_swagger_object(self, field, swagger_object_type, use_references, **kw):
        # pylint: disable=unused-variable,invalid-name
        if not isinstance(field, fields.RatingField):
            return NotHandled

        SwaggerType, ChildSwaggerType = self._get_partial_types(
            field, swagger_object_type, use_references, **kw
        )
        kwargs = {
            'type': openapi.TYPE_NUMBER,
            'format': FORMAT_RATING,
            X_OPTIONS: {
                'min_value': field.min_value,
                'max_value': field.max_value,
                'step': field.step,
                'style': field.front_style,
                'color': field.color,
                'fa_class': field.fa_class
            }
        }
        return SwaggerType(**field_extra_handler(field, **kwargs))


class NamedBinaryImageInJsonFieldInspector(FieldInspector):
    def field_to_swagger_object(self, field, swagger_object_type, use_references, **kw):
        # pylint: disable=unused-variable,invalid-name
        if isinstance(field, fields.NamedBinaryImageInJsonField):
            kwargs = items = deepcopy(basic_type_info[fields.NamedBinaryFileInJsonField])
        elif isinstance(field, fields.MultipleNamedBinaryImageInJsonField):
            kwargs = deepcopy(basic_type_info[fields.MultipleNamedBinaryFileInJsonField])
            items = kwargs['items']
        else:
            return NotHandled

        SwaggerType, ChildSwaggerType = self._get_partial_types(
            field, swagger_object_type, use_references, **kw
        )

        items['x-format'] = FORMAT_NAMED_BIN_IMAGE
        x_validators = items['x-validators'] = {
            'extensions': set()
        }
        for validator in filter(lambda x: isinstance(x, validators.ImageBaseSizeValidator), field.validators):
            for orientation in validator.orientation:
                for size_type, default_size in (('min', 1), ('max', float('inf'))):
                    size_name = f'{size_type}_{orientation}'
                    img_size = getattr(validator, size_name, default_size)
                    if size_name in x_validators:
                        if size_type == 'min' and x_validators[size_name] > img_size:
                            continue
                        elif size_type == 'max' and x_validators[size_name] < img_size:
                            continue
                    x_validators[size_name] = img_size
            if x_validators['extensions']:
                x_validators['extensions'] = x_validators['extensions'].intersection(validator.extensions)
            else:
                x_validators['extensions'] = set(validator.extensions)

        x_validators['extensions'] = tuple(x_validators['extensions'])
        if not x_validators['extensions']:
            del x_validators['extensions']

        return SwaggerType(**field_extra_handler(field, **kwargs))


class MaskedFieldInspector(FieldInspector):
    def field_to_swagger_object(self, field, swagger_object_type, use_references, **kw):
        # pylint: disable=unused-variable,invalid-name
        if not isinstance(field, fields.MaskedField):
            return NotHandled

        SwaggerType, ChildSwaggerType = self._get_partial_types(
            field, swagger_object_type, use_references, **kw
        )
        kwargs = {
            'type': openapi.TYPE_STRING,
            'format': FORMAT_MASKED,
            X_OPTIONS: {
                'mask': field.mask
            }
        }
        return SwaggerType(**field_extra_handler(field, **kwargs))


class CSVFileFieldInspector(FieldInspector):
    def field_to_swagger_object(self, field, swagger_object_type, use_references, **kw):
        if not isinstance(field, fields.CSVFileField):
            return NotHandled
        # pylint: disable=unused-variable,invalid-name
        SwaggerType, ChildSwaggerType = self._get_partial_types(
            field, swagger_object_type, use_references, **kw
        )

        kwargs = {
            'type': openapi.TYPE_STRING,
            'format': FORMAT_CSV_FILE,
            X_OPTIONS: {
                'delimiter': field.delimiter,
                'minColumnWidth': field.min_column_width,
                'items': self.probe_field_inspectors(field.items, swagger_object_type, False)
            }
        }

        return SwaggerType(**field_extra_handler(field, **kwargs))


class DecimalFieldInspector(FieldInspector):
    def field_to_swagger_object(self, field, swagger_object_type, use_references, **kw):
        # pylint: disable=unused-variable,invalid-name
        if not isinstance(field, DecimalField):
            return NotHandled

        SwaggerType, ChildSwaggerType = self._get_partial_types(
            field, swagger_object_type, use_references, **kw
        )
        kwargs = {
            'type': decimal_field_type(field),
            'format': FORMAT_DECIMAL,
            X_OPTIONS: {
                'decimal_places': field.decimal_places,
                'max_digits': field.max_digits
            }
        }

        if field.default != empty:
            kwargs['default'] = str(field.default)

        return SwaggerType(**field_extra_handler(field, **kwargs))


class NestedFilterInspector(CoreAPICompatInspector):
    def get_filter_parameters(self, filter_backend):  # nocv
        subaction_list_actions = [
            f'{name}_list'
            for name in getattr(self.view, '_nested_args', {}).keys()
        ]
        if self.view.action not in subaction_list_actions:
            return NotHandled
        if self.method != 'GET':
            return NotHandled
        nested_view = getattr(self.view, self.view.action, None)
        nested_view_filter_class = getattr(nested_view, '_nested_filter_class', None)
        filter_class = getattr(self.view, 'filterset_class', getattr(self.view, 'filter_class', None))
        self.view.filter_class = nested_view_filter_class
        result = super().get_filter_parameters(filter_backend)
        self.view.filter_class = filter_class
        return result


class ArrayFilterQueryInspector(CoreAPICompatInspector):
    @cached_property
    def fields_map(self):
        return {
            f.name: f
            for f in self.view.get_queryset().model._meta.fields
        }

    def coreapi_field_to_parameter(self, field, schema=None):
        """
        Convert an instance of `coreapi.Field` to a swagger :class:`.Parameter` object.

        :param coreapi.Field field:
        :param coreschema..Schema schema:
        :rtype: openapi.Parameter
        """
        location_to_in = {
            'query': openapi.IN_QUERY,
            'path': openapi.IN_PATH,
            'form': openapi.IN_FORM,
            'body': openapi.IN_FORM,
        }
        coreapi_types = {
            coreschema.Integer: openapi.TYPE_INTEGER,
            coreschema.Number: openapi.TYPE_NUMBER,
            coreschema.String: openapi.TYPE_STRING,
            coreschema.Boolean: openapi.TYPE_BOOLEAN,
            coreschema.Array: openapi.TYPE_ARRAY,
        }

        coreschema_attrs = ['format', 'pattern', 'enum', 'min_length', 'max_length']
        schema_field = schema or field.schema
        attributes = {}
        if isinstance(schema_field, coreschema.Array):
            attributes['collectionFormat'] = 'csv'
            param = self.coreapi_field_to_parameter(field, schema_field.items)
            attributes['items'] = openapi.Items(**OrderedDict(
                (attr, getattr(param, attr, None))
                for attr in coreschema_attrs + ['type']
            ))
            attributes['minItems'] = schema_field.min_items
            attributes['maxItems'] = schema_field.max_items
            attributes['uniqueItems'] = schema_field.unique_items
            coreschema_attrs = ()

        schema_type = coreapi_types.get(type(schema_field), openapi.TYPE_STRING)
        if schema is not None and \
           field.name in ('id', 'id__not') and \
           isinstance(self.fields_map.get(field.name.split('__')[0]), (models.AutoField, models.IntegerField)):
            schema_type = openapi.TYPE_INTEGER

        return openapi.Parameter(
            name=field.name,
            in_=location_to_in[field.location],
            required=field.required,
            description=force_real_str(schema_field.description) if schema_field else None,
            type=schema_type,
            **OrderedDict((attr, getattr(schema_field, attr, None)) for attr in coreschema_attrs),
            **attributes
        )


class VSTReferencingSerializerInspector(ReferencingSerializerInspector):
    def get_serializer_ref_name(self, serializer: Any):
        if isinstance(serializer, serializers.serializers.ListSerializer):
            return super().get_serializer_ref_name(serializer.child)
        return super().get_serializer_ref_name(serializer)

    def handle_schema(self, field: Any):
        ref_name = self.get_serializer_ref_name(field)
        definitions = self.components.with_scope(openapi.SCHEMA_DEFINITIONS)

        if ref_name not in definitions:
            return

        schema = definitions[ref_name]

        if getattr(schema, '_handled', False):
            return

        schema_properties = tuple(schema['properties'].keys())
        serializer_class = schema._NP_serializer  # pylint: disable=protected-access

        non_bulk_methods = getattr(serializer_class, '_non_bulk_methods', None)
        schema_properties_groups = OrderedDict(
            getattr(serializer_class, '_schema_properties_groups', None) or {'': schema_properties}
        )
        translate_model = getattr(serializer_class, '_translate_model', None)

        view_field_name = getattr(serializer_class, '_view_field_name', None)

        if view_field_name is None and schema_properties:
            view_field_name = get_first_match_name(schema_properties, schema_properties[0])

        if schema_properties_groups:
            not_handled = set(schema['properties']) - set(_get_handled_props(schema_properties_groups))

            if not_handled:
                schema_properties_groups[''] = [
                    prop
                    for prop in schema['properties']
                    if prop in not_handled
                ]
        schema['x-properties-groups'] = schema_properties_groups
        schema['x-view-field-name'] = view_field_name

        if non_bulk_methods:
            schema['x-non-bulk-methods'] = non_bulk_methods

        if translate_model:
            schema['x-translate-model'] = translate_model

        schema._handled = True  # pylint: disable=protected-access

    def field_to_swagger_object(self, field: Any, swagger_object_type: Any, use_references: Any, **kwargs: Any):
        if isinstance(field, FileResponse):
            return openapi.Schema(type='file')

        result = super().field_to_swagger_object(field, swagger_object_type, use_references, **kwargs)

        if result != NotHandled:
            self.handle_schema(field)

        return result



import typing as _t

from django.utils.encoding import force_str
from django.db import models
from rest_framework.filters import BaseFilterBackend, OrderingFilter
from django_filters.rest_framework.backends import DjangoFilterBackend as BaseDjangoFilterBackend
from django_filters import compat, filters
from vstutils.utils import raise_context, translate as _

from .filters import extra_filter


class DjangoFilterBackend(BaseDjangoFilterBackend):
    def get_coreschema_field(self, field):
        kwargs = {
            'description': str(field.extra.get('help_text', '')),
        }
        if isinstance(field, filters.NumberFilter):
            field_cls = compat.coreschema.Number
        elif isinstance(field, filters.BooleanFilter):
            field_cls = compat.coreschema.Boolean
        elif isinstance(field, filters.ChoiceFilter):
            field_cls = compat.coreschema.Enum
            kwargs['enum'] = tuple(dict(field.field.choices).keys())
        else:
            field_cls = compat.coreschema.String
        if field.method == extra_filter:  # pylint: disable=comparison-with-callable
            result = compat.coreschema.Array(
                items=field_cls(),
                min_items=1,
                unique_items=True,
                **kwargs
            )
        else:
            result = field_cls(**kwargs)
        return result


class OrderingFilterBackend(OrderingFilter):
    def _get_fields_for_schema(self, view):
        for field in self.get_valid_fields(view.get_queryset(), view, {'request': view.request}):
            yield field[0]
            yield f'-{field[0]}'

    def get_schema_fields(self, view):
        try:
            fields = tuple(self._get_fields_for_schema(view))
        except:
            return super().get_schema_fields(view)
        return [
            compat.coreapi.Field(
                name=self.ordering_param,
                required=False,
                location='query',
                schema=compat.coreschema.Array(
                    title=_(force_str(self.ordering_title)),
                    description=_(force_str(self.ordering_description)),
                    items=compat.coreschema.Enum(enum=fields),
                    min_items=1,
                    unique_items=True,
                )
            )
        ]


# Call standard filtering
class VSTFilterBackend(BaseFilterBackend):
    """
    A base filter backend class to be inherited from.
    Example:

        .. sourcecode:: python

            from django.utils import timezone
            from django.db.models import Value, DateTimeField

            from vstutils.api.filter_backends import VSTFilterBackend

            class CurrentTimeFilterBackend(VSTFilterBackend):
                def filter_queryset(self, request, queryset, view):
                    return queryset.annotate(current_time=Value(timezone.now(), output_field=DateTimeField()))

        In this example Filter Backend annotates time in current timezone to any connected
        model's queryset.

    In some cases it may be necessary to provide a parameter from a query of request.
    To define this parameter in the schema, you must overload the get_schema_fields
    function and specify a list of parameters to use.

    Example:

        .. sourcecode:: python

            from django.utils import timezone
            from django.db.models import Value, DateTimeField

            from vstutils.api.filter_backends import VSTFilterBackend

            class ConstantCurrentTimeForQueryFilterBackend(VSTFilterBackend):
                query_param = 'constant'

                def filter_queryset(self, request, queryset, view):
                    if self.query_param in request.query_params and request.query_params[self.query_param]:
                        queryset = queryset.annotate(**{
                            request.query_params[self.query_param]: Value(timezone.now(), output_field=DateTimeField())
                        })
                    return queryset

                    def get_schema_fields(self, view):
                        return [
                            compat.coreapi.Field(
                                name=self.query_param,
                                required=False,
                                location='query',
                                schema=compat.coreschema.String(
                                    description='Annotate value to queryset'
                                ),
                            )
                        ]

        In this example Filter Backend annotates time in current timezone to any connected
        model's queryset with field name from query `constant`.
    """
    __slots__ = ()
    required = False

    def filter_queryset(self, request, queryset, view):
        raise NotImplementedError  # nocv

    def get_schema_fields(self, view):
        """
        You can also make the filter controls available to the schema autogeneration that REST framework provides,
        by implementing this method. The method should return a list of coreapi.Field instances.
        """
        # pylint: disable=unused-argument
        return []


class HideHiddenFilterBackend(VSTFilterBackend):
    """Filter Backend that hides all objects with hidden=True from the queryset"""
    __slots__ = ()
    required = True

    def filter_queryset(self, request, queryset, view):
        # pylint: disable=unused-argument
        """
        Clear objects with hidden attr from queryset.
        """
        return getattr(queryset, 'cleared', queryset.all)()


class SelectRelatedFilterBackend(VSTFilterBackend):
    """
    Filter Backend that will automatically call prefetch_related and select_related on all relations in queryset.
    """
    __slots__ = ()
    required = True
    fields_fetch_map = {
        'select': (models.ForeignKey,),
        'prefetch': (models.ManyToManyField, models.ManyToManyField.rel_class)
    }

    def filter_model_fields(self, view, field_types):
        return tuple(
            f.name
            for f in view.get_serializer_class().Meta.model()._meta.fields
            if isinstance(f, field_types)
        )

    def filter_by_func(self, queryset, queryset_func_name, related):
        if related:
            return getattr(queryset, queryset_func_name)(*related)
        return queryset

    def prefetch(self, queryset_func_name, view, queryset):
        with raise_context():
            queryset = self.filter_by_func(
                queryset,
                f'{queryset_func_name}_related',
                self.filter_model_fields(view, self.fields_fetch_map[queryset_func_name])
            )
        return queryset

    def filter_queryset(self, request, queryset, view):
        """
        Select+prefetch related in queryset.
        """
        if request.method != 'GET':
            return queryset

        if not queryset.query.select_related and getattr(view, 'select_related', True):
            queryset = self.prefetch('select', view, queryset)

        # pylint: disable=protected-access
        if not queryset._prefetch_related_lookups and getattr(view, 'prefetch_related', True):
            queryset = self.prefetch('prefetch', view, queryset)
        return queryset


class DeepViewFilterBackend(VSTFilterBackend):
    """
    Backend that filters queryset by column from `deep_parent_field` property of the model.
    Value for filtering must be provided in query param `__deep_parent`.

    If param is missing then no filtering is applied.

    If param is empty value (`/?__deep_parent=`) then objects with no parent (the value of the field whose name is
    stored in the property `deep_parent_field` of the model is None) returned.

    This filter backend and nested view is automatically added when model has `deep_parent_field` property.

    Example:
        .. sourcecode:: python

            from django.db import models
            from vstutils.models import BModel

            class DeepNestedModel(BModel):
                name = models.CharField(max_length=10)
                parent = models.ForeignKey('self', null=True, default=None, on_delete=models.CASCADE)

                deep_parent_field = 'parent'
                deep_parent_allow_append = True

                class Meta:
                    default_related_name = 'deepnested'

    In example above if we add this model under path '`deep`', following views will be created: `/deep/` and
    `/deep/{id}/deepnested/`.

    Filter backend can be used as `/deep/?__deep_parent=1` and will return all `DeepNestedModel` objects
    whose parent's primary key is `1`.

    You can also use generic DRF views, for that you still must set `deep_parent_field`
    to your model and manually add `DeepViewFilterBackend` to
    `filter_backends <https://www.django-rest-framework.org/api-guide/filtering/#djangofilterbackend>`_ list.
    """
    field_name = '__deep_parent'
    field_types = {
        compat.coreschema.Integer: filters.NumberFilter,  # type: ignore
        compat.coreschema.String: filters.CharFilter,  # type: ignore
    }

    def filter_queryset(self, request, queryset, view):
        model = queryset.model
        parent_name: _t.Optional[_t.Text] = getattr(model, 'deep_parent_field', None)
        nested = getattr(view, 'nested_id', False) or getattr(getattr(view, 'nested_parent_object', None), 'id', None)
        if view.action != 'list':
            if parent_name and nested:
                return queryset.get_children(with_current=True)
            return queryset
        if not parent_name or self.field_name not in request.query_params:
            return queryset
        filter_type_class = self.field_types[type(self.get_coreschema_field(model))]
        filter_data = request.query_params.get(self.field_name)
        if not filter_data:
            return queryset.filter(**{f'{parent_name}__isnull': True})
        pk_name = model._meta.pk.attname
        parent_qs = model.objects.filter(**{pk_name: filter_data})
        parent_qs = getattr(parent_qs, 'cleared', parent_qs.all)()
        if nested:
            queryset = model.objects.all()
        return filter_type_class(field_name=parent_name, lookup_expr='in').filter(
            queryset,
            value=parent_qs.values(pk_name)
        )

    def get_coreschema_field(self, model: models.Model):
        primary_key_field: _t.Optional[models.Field] = model._meta.pk
        if isinstance(primary_key_field, models.IntegerField):
            field_cls = compat.coreschema.Integer  # type: ignore
        else:  # nocv
            field_cls = compat.coreschema.String  # type: ignore
        return field_cls(
            description=''
        )

    def get_schema_fields(self, view):
        return [
            compat.coreapi.Field(
                name=self.field_name,
                required=False,
                location='query',
                schema=self.get_coreschema_field(view.get_queryset().model),
            )
        ]



import typing as _t
import logging
import traceback
import functools
from concurrent.futures import ThreadPoolExecutor, Executor
from collections import OrderedDict

import orjson
from django.conf import settings
from django.db import transaction
from django.http import HttpResponse, HttpRequest
from django.contrib.auth.models import AbstractUser
from django.test.client import Client, ClientHandler
from django.test.utils import modify_settings
from drf_yasg.views import SPEC_RENDERERS
from rest_framework import serializers, views, versioning, request as drf_request
from rest_framework.authentication import (
    SessionAuthentication,
    BasicAuthentication,
    TokenAuthentication,
    BaseAuthentication
)

from . import responses
from .decorators import cache_method_result
from .serializers import DataSerializer
from .validators import UrlQueryStringValidator
from .renderers import ORJSONRenderer
from ..utils import Dict, raise_context, patch_gzip_response
from ..middleware import BaseMiddleware

RequestType = _t.Union[drf_request.Request, HttpRequest]
logger: logging.Logger = logging.getLogger('vstutils')

THREADS_COUNT = settings.BULK_THREADS
API_URL: _t.Text = settings.API_URL
DEFAULT_VERSION = settings.VST_API_VERSION
REST_METHODS: _t.List[_t.Text] = [
    m.upper() for m in views.APIView.http_method_names
]

default_authentication_classes = (
    SessionAuthentication,
    BasicAuthentication,
    TokenAuthentication
)

append_to_list = list.append
response_headers_to_pass = (
    "ETag",
    "Location",
    "Pagination-Identifiers",
)


@functools.singledispatch
def _get_request_data(request_data: _t.Iterable) -> _t.Union[_t.List, _t.Tuple]:
    assert isinstance(request_data, (list, tuple)), 'Request data must be list or tuple.'
    return request_data


@_get_request_data.register(dict)
def _get_request_data_dict(request_data):
    return [request_data]


@_get_request_data.register(str)
def _get_request_data_str(request_data):
    return _get_request_data(orjson.loads(request_data))  # nocv


def _iter_request(request, operation_handler, context):
    executor_class = _DummyExecutor
    if request.method not in ('POST', 'PUT'):
        executor_class = ThreadPoolExecutor if THREADS_COUNT else executor_class
    handler = lambda o: operation_handler(o, context)
    with executor_class(max_workers=THREADS_COUNT) as executor:
        for operation_result in executor.map(handler, _get_request_data(request.data)):
            yield operation_result


def _join_paths(*args) -> _t.Text:
    """Join multiple path fragments into one

    :param *args: List of items that can be anything like '/a/b/c', 'b/c/', 1, 'v1'
    :returns: Path that starts and ends with
    """
    return f"/{'/'.join(str(arg).strip('/') for arg in args)}/"


class _DummyExecutor(Executor):
    # pylint: disable=abstract-method

    def __init__(self, *args, **kwargs):
        super().__init__()

    def map(self, fn, *iterables, timeout=None, chunksize=1):
        return map(fn, *iterables)


class ParseResponseDict(dict):
    __slots__ = ('timing',)
    timing: _t.SupportsFloat

    def __init__(self, path: _t.Text, method: _t.Text, response: HttpResponse):
        super().__init__(
            path=path,
            status=response.status_code,
            data=self._get_rendered(response),
            method=method,
            headers={
                header: response.headers[header]
                for header in response.headers
                if header in response_headers_to_pass
            }
        )
        self.timing = float(response.get('Response-Time', '0.0'))

    def _get_rendered(self, response: _t.Union[HttpResponse, responses.BaseResponseClass]):
        with raise_context():
            result = response.data  # type: ignore
            if isinstance(result, dict):
                return Dict(result)
        with raise_context():
            if isinstance(response.accepted_renderer, ORJSONRenderer) and response.is_rendered:  # type: ignore
                return response.rendered_content  # type: ignore
        if response.status_code != 404 and getattr(response, "rendered_content", False):  # nocv
            return orjson.loads(response.rendered_content.decode())  # type: ignore
        return Dict(detail=str(response.content.decode('utf-8')))


class BulkRequestType(drf_request.Request, HttpRequest):  # type: ignore
    # pylint: disable=abstract-method
    data: _t.List[_t.Dict[_t.Text, _t.Any]]  # type: ignore
    version: _t.Optional[_t.Text]
    successful_authenticator: _t.Optional[BaseAuthentication]


class BulkMiddleware(BaseMiddleware):
    __slots__ = ()

    def request_handler(self, request: HttpRequest) -> HttpRequest:
        request.is_bulk = True  # type: ignore
        if 'user' in request.META:
            request.user = request.META.pop('user')
            # pylint: disable=protected-access
            request._cached_user = request.user  # type: ignore
        if 'language' in request.META:
            request.language = request.META.pop('language')  # type: ignore
        if 'session' in request.META:
            request.session = request.META.pop('session')  # type: ignore
        return request


class BulkClientHandler(ClientHandler):
    __slots__ = ()

    @modify_settings(MIDDLEWARE=settings.MIDDLEWARE_ENDPOINT_CONTROL)
    def __init__(self, *args, **kwargs):
        super().__init__(enforce_csrf_checks=False, *args, **kwargs)
        if self.__class__.__name__ == 'BulkClientHandler':
            self.load_middleware()


class BulkClient(Client):
    __slots__ = ('user', 'language', 'session', 'exc_info')
    handler: BulkClientHandler = BulkClientHandler()
    user: _t.Optional[AbstractUser]

    def __init__(self, enforce_csrf_checks=False, **defaults):
        # pylint: disable=bad-super-call
        self.user = defaults.pop('user', None)
        self.language = defaults.pop('language', None)
        self.session = defaults.pop('session', None)
        super(Client, self).__init__(**defaults)
        self.exc_info = None

    def request(self, **request):
        if self.user:
            request['user'] = self.user
        if self.language:
            request['language'] = self.language
        if self.session:
            request['session'] = self.session
        response = self.handler(self._base_environ(**request))
        if response.cookies:
            self.cookies.update(response.cookies)
        return response


class FormatDataFieldMixin:
    """
    Mixin for fields that can format "<< >>" templates inside strings
    """
    __slots__ = ()
    requires_context: bool = True
    context: _t.Dict

    def to_internal_value(self, data) -> _t.Text:
        result = super().to_internal_value(data)  # type: ignore

        if isinstance(result, str) \
                and '<<' in result \
                and '>>' in result \
                and not ('{' in result and '}' in result) \
                and 'results' in self.context:
            result = result.replace('<<', '{').replace('>>', '}').format(
                *self.context['results'],
                **self.context['variables'],
            )
            with raise_context():
                return orjson.loads(result)

        return result


class TemplateStringField(FormatDataFieldMixin, serializers.CharField):
    """
    Field that can format "<< >>" templates inside strings
    """
    __slots__ = ()


class RequestDataField(FormatDataFieldMixin, DataSerializer):
    """
    Field that can handle basic data types and recursise
    format template strings inside them
    """
    __slots__ = ()

    def to_internal_value(self, data):
        if isinstance(data, str):
            pass

        elif isinstance(data, (list, tuple)):
            return [self.to_internal_value(i) for i in data]

        elif isinstance(data, (dict, OrderedDict)):
            return type(data)(
                (super(RequestDataField, self).to_internal_value(k), self.to_internal_value(v))
                for k, v in data.items()
            )

        return super(RequestDataField, self).to_internal_value(data)


class MethodChoicesField(serializers.ChoiceField):
    """Field for HTTP method"""
    __slots__ = ()

    def __init__(self, choices: _t.List = None, **kwargs):
        super().__init__(choices or REST_METHODS, **kwargs)

    def to_internal_value(self, data):
        return super(MethodChoicesField, self).to_internal_value(str(data).upper())


class PathField(TemplateStringField):
    __slots__ = ()

    def to_internal_value(self, data):

        if isinstance(data, str):
            data = (data, )

        return _join_paths(*[
            super(PathField, self).to_internal_value(path)
            for path in data
        ])


class OperationSerializer(serializers.Serializer):
    # pylint: disable=abstract-method
    __slots__ = ()
    renderer = ORJSONRenderer()

    path = PathField(required=True)
    method = MethodChoicesField(required=True)
    headers = serializers.DictField(child=TemplateStringField(), default={})
    data = RequestDataField(required=False, default=None, allow_null=True)  # type: ignore
    status = serializers.IntegerField(read_only=True, default=500)
    info = serializers.CharField(read_only=True)
    query = TemplateStringField(required=False,
                                allow_blank=True,
                                validators=[UrlQueryStringValidator()],
                                write_only=True)
    let = TemplateStringField(required=False,
                              write_only=True)
    version = serializers.ChoiceField(choices=list(settings.API.keys()),
                                      default=settings.VST_API_VERSION,
                                      write_only=True)

    def to_representation(self, instance: _t.Dict[_t.Text, _t.Any]) -> Dict:
        return Dict(super().to_representation(instance))

    def get_operation_method(self, method: _t.Text) -> _t.Callable:
        return getattr(self.context.get('client'), method.lower())

    def create(self, validated_data: _t.Dict[_t.Text, _t.Union[_t.Text, _t.Mapping]]) -> ParseResponseDict:
        # pylint: disable=protected-access
        method_name = str(validated_data['method']).lower()
        method = self.get_operation_method(method_name)
        url = _join_paths(API_URL, validated_data['version'], validated_data['path'])
        if 'query' in validated_data and validated_data['query']:
            url += '?' + str(validated_data['query'])
        if method_name != 'get':
            method = transaction.atomic()(method)
        data = validated_data['data']
        if data and method_name != 'get':
            data = self.renderer.render(data, media_type=self.renderer.media_type)
        result = ParseResponseDict(
            path=url,
            method=method_name,
            response=method(  # type: ignore
                url,
                content_type='application/json',
                secure=self.context['request']._request.is_secure(),
                data=data,
                **validated_data['headers']
            )
        )
        if 'let' in validated_data:
            self.context['variables'][validated_data['let']] = result
        return result


class EndpointViewSet(views.APIView):
    """
    Default API-endpoint viewset.
    """
    __slots__ = ('results',)

    throttle_classes = []  # type: ignore
    schema = None  # type: ignore
    versioning_class = versioning.QueryParameterVersioning  # type: ignore
    renderer_classes = list(views.APIView.renderer_classes) + list(SPEC_RENDERERS)
    session_cookie_name: _t.ClassVar[_t.Text] = settings.SESSION_COOKIE_NAME
    client_environ_keys_copy: _t.List[_t.Text] = [
        "SCRIPT_NAME",
        "SERVER_NAME",
        "SERVER_PORT",
        "SERVER_PROTOCOL",
        "SERVER_SOFTWARE",
        "REMOTE_ADDR",
        settings.SECURE_PROXY_SSL_HEADER[0],
        "HTTP_HOST",
        "HTTP_USER_AGENT"
    ]

    #: One operation serializer class.
    serializer_class: _t.ClassVar[_t.Type[OperationSerializer]] = OperationSerializer

    def get_client(self, request: BulkRequestType) -> BulkClient:
        """
        Returns test client and guarantees that if bulk request comes
        authenticated than test client will be authenticated with the same user
        """
        return BulkClient(**self.original_environ_data(request=request))

    def original_environ_data(self, request: BulkRequestType, *args) -> _t.Dict:
        get_environ = request.META.get
        kwargs = {}
        for env_var in tuple(self.client_environ_keys_copy) + args:
            value = get_environ(env_var, None)
            if value:
                kwargs[env_var] = str(value)
        if request.user.is_authenticated:
            if isinstance(request.successful_authenticator, SessionAuthentication):
                kwargs['HTTP_COOKIE'] = str(request.META.get('HTTP_COOKIE'))
            elif isinstance(request.successful_authenticator, (BasicAuthentication, TokenAuthentication)):
                kwargs['HTTP_AUTHORIZATION'] = str(request.META.get('HTTP_AUTHORIZATION'))
            kwargs['user'] = request.user  # type: ignore
        kwargs['language'] = getattr(request, 'language', None)
        kwargs['session'] = getattr(request, 'session', None)
        return kwargs

    def get_serializer(self, *args, **kwargs) -> OperationSerializer:
        """
        Return the serializer instance that should be used for validating and
        deserializing input, and for serializing output.
        """
        serializer_class = self.get_serializer_class()
        kwargs['context'] = self.get_serializer_context(kwargs.get('context', {}))
        return serializer_class(*args, **kwargs)

    @cache_method_result
    def get_serializer_class(self):
        """
        Return the class to use for the serializer.
        Defaults to using `self.serializer_class`.

        You may want to override this if you need to provide different
        serializations depending on the incoming request.

        (Eg. admins get full serialization, others get basic serialization)
        """
        assert self.serializer_class is not None, (
            "'%s' should either include a `serializer_class` attribute, "
            "or override the `get_serializer_class()` method."
            % self.__class__.__name__
        )

        return self.serializer_class

    def get_serializer_context(self, context) -> dict:
        """
        Extra context provided to the serializer class.
        """
        if 'client' not in context:  # nocv
            context = context.copy()
            context['client'] = self.get_client(_t.cast(BulkRequestType, self.request))
        return {
            'request': self.request,
            'view': self,
            **context
        }

    def operate(self, operation_data: _t.Dict, context: _t.Dict) -> _t.Tuple[_t.Dict, _t.SupportsFloat]:
        """Method used to handle one operation and return result of it"""
        serializer = self.get_serializer(data=operation_data, context=context)
        try:
            serializer.is_valid(raise_exception=True)
            return serializer.to_representation(serializer.save()), serializer.instance.timing  # type: ignore
        except Exception as err:
            return {
                'path': 'bulk',
                'info': {
                    'errors': getattr(serializer, '_errors', traceback.format_exc()),
                    'operation_data': operation_data
                },
                'status': 500,
                'data': {'detail': f'Error in bulk request data. See info. Original message: {str(err)}'}
            }, 0.0

    def get(self, request: BulkRequestType) -> HttpResponse:
        """Returns response with swagger ui or openapi json schema if ?format=openapi"""

        url = f'/api/{getattr(request, "version", DEFAULT_VERSION) or DEFAULT_VERSION}/_openapi/'

        if request.query_params.get('format') == 'openapi':  # type: ignore
            url += '?format=openapi'
            should_gzip = True
        else:
            should_gzip = False

        response = self.get_client(request).get(url, secure=request.is_secure())

        if should_gzip:
            patch_gzip_response(response, request)

        return response

    def post(self, request: BulkRequestType) -> responses.BaseResponseClass:
        """Execute transactional bulk request"""
        try:
            with transaction.atomic():
                return self.put(request, allow_fail=False)
        except Exception:
            logger.debug(traceback.format_exc())
            return responses.HTTP_502_BAD_GATEWAY(self.results)

    def put(self, request: BulkRequestType, allow_fail=True) -> responses.BaseResponseClass:
        """Execute non transaction bulk request"""
        context: _t.Dict[_t.Text, _t.Union[_t.List, _t.Dict, BulkClient]] = {
            'client': self.get_client(request),
            'results': self.results,
            'variables': {},
        }
        timings: _t.List = []
        for result, timing in _iter_request(request, self.operate, context):
            append_to_list(self.results, result)
            append_to_list(timings, timing)
            if not allow_fail and not (100 <= result.get('status', 500) < 400):
                raise Exception(f'Execute transaction stopped. Error message: {str(result)}')
        response = responses.HTTP_200_OK(self.results, timings={f'op{i}': float(j) for i, j in enumerate(timings)})
        for cookie_name, cookie_value in context['client'].cookies.items():  # type: ignore
            if cookie_value.value != request.COOKIES.get(cookie_name, None):
                response.cookies[cookie_name] = cookie_value
        return response

    def patch(self, request: BulkRequestType) -> responses.BaseResponseClass:
        return self.put(request)

    def initial(self, request: drf_request.Request, *args, **kwargs):
        super().initial(request, *args, **kwargs)
        self.results: _t.List[_t.Dict[_t.Text, _t.Any]] = []

    def finalize_response(self, request: drf_request.Request, *args, **kwargs):
        if not isinstance(request.successful_authenticator, default_authentication_classes):
            self.get_client(_t.cast(BulkRequestType, self.request)).logout()
        return super().finalize_response(request, *args, **kwargs)
